# -*- coding: utf-8 -*-

import logging
import time

import gevent

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.Command.Tango import DeviceProxy
from mxcubecore.TaskUtils import task


class EnvironmentPhase:
    TRANSFER = 0
    CENTRING = 1
    COLLECT = 2
    DEFAULT = 3
    BEAMVIEW = 4
    FLUOX = 5
    MANUALTRANSFER = 6
    INPROGRESS = 7
    VISUSAMPLE = 8

    phasedesc = {
        "TRANSFER": TRANSFER,
        "CENTRING": CENTRING,
        "COLLECT": COLLECT,
        "DEFAULT": DEFAULT,
        "BEAMVIEW": BEAMVIEW,
        "FLUOX": FLUOX,
        "MANUALTRANSFER": MANUALTRANSFER,
        "INPROGRESS": INPROGRESS,
        "VISUSAMPLE": VISUSAMPLE,
    }

    @staticmethod
    def phase(phasename):
        return EnvironmentPhase.phasedesc.get(phasename, None)


class EnvironemntState:
    UNKNOWN, ON, RUNNING, ALARM, FAULT = (0, 1, 10, 13, 14)  # Like PyTango stated

    # TangoStates = {
    #    Unknown     = 0
    #    On          = 1
    #    Loaded      = 2
    #    Loading     = 3
    #    Unloading   = 4
    #    Selecting   = 5
    #    Scanning    = 6
    #    Resetting   = 7
    #    Charging    = 8
    #    Moving      = 9
    #    Running     = 10
    #    StandBy     = 11
    #    Disabled    = 12
    #    Alarm       = 13
    #    Fault       = 14
    #    Initializing= 15
    #    Closing     = 16
    #    Off         = 17
    # }

    statedesc = {ON: "ON", RUNNING: "RUNNING", ALARM: "ALARM", FAULT: "FAULT"}

    @staticmethod
    def tostring(state):
        return SampleChangerState.statedesc.get(state, "Unknown")


class PX1Environment(HardwareObject):
    def __init__(self, name):
        super().__init__(name)
        self.auth = None
        self.device = None

    def init(self):
        self.device = DeviceProxy(self.get_property("tangoname"))

        try:
            self.state_chan = self.get_channel_object("State")
            self.state_chan.connect_signal("update", self.stateChanged)

        except KeyError:
            logging.getLogger().warning("%s: cannot report State", self.name())

        try:
            self.chanAuth = self.get_channel_object("beamlineMvtAuthorized")
            self.chanAuth.connect_signal("update", self.setAuthorizationFlag)
            # state = self.state_chan.get_value()

        except KeyError:
            logging.getLogger().warning("%s: cannot report State", self.name())

        try:
            self.usingCapillaryChannel = self.get_channel_object("usingCapillary")
            self.setUsingCapillary(True)
        except Exception:
            self.usingCapillaryChannel = None

        try:
            self.beamstopPositionChannel = self.get_channel_object("beamstopPosition")
        except Exception:
            self.beamstopPositionChannel = None

        if self.device is not None:
            self.set_is_ready(True)

            self.cmds = {
                EnvironmentPhase.TRANSFER: self.device.GoToTransfertPhase,
                EnvironmentPhase.CENTRING: self.device.GoToCentringPhase,
                EnvironmentPhase.COLLECT: self.device.GoToCollectPhase,
                EnvironmentPhase.DEFAULT: self.device.GoToDefaultPhase,
                # EnvironmentPhase.BEAMVIEW: self.device.GoToBeamViewPhase,
                EnvironmentPhase.FLUOX: self.device.GoToFluoXPhase,
                EnvironmentPhase.MANUALTRANSFER: self.device.GoToManualTransfertPhase,
                EnvironmentPhase.VISUSAMPLE: self.device.GoToVisuSamplePhase,
            }

    # ---- begin state handling
    #
    def stateChanged(self, value):
        self.emit("StateChanged", (value,))

    def get_state(self):
        state = str(self.state_chan.get_value())
        return state

    def isBusy(self, timeout=None):
        state = self.stateChan.get_value()
        return state not in [EnvironmentState.ON]

    def wait_ready(self, timeout=None):
        self._waitState(["ON"], timeout)

    def _waitState(self, states, timeout=None):
        if self.device is None:
            return

        _debut = time.time()
        with gevent.Timeout(timeout, Exception("Timeout waiting for device ready")):
            waiting = True
            while waiting:
                state = self.readState()
                if state in states:
                    waiting = False
                gevent.sleep(0.05)

    #
    # ------- end state handling
    # logging.debug("PX1environment: end _waitState in %.1f sec" % (time.time() - _debut))

    # ------- begin phase handling
    #
    def isPhaseTransfer(self):
        return self.device.readyForTransfert

    def isPhaseCollect(self):
        return self.readyForCollect()

    def isPhaseVisuSample(self):
        return self.device.readyForVisuSample

    def isPhaseFluoScan(self):
        return self.device.readyForFluoScan

    def readyForCentring(self):
        if self.device is not None:
            return self.device.readyForCentring
        else:
            return None

    def readyForCollect(self):
        if self.device is not None:
            return self.device.readyForCollect
        else:
            return None

    def readyForDefaultPosition(self):
        if self.device is not None:
            return self.device.readyForDefaultPosition
        else:
            return None

    def readyForFluoScan(self):
        if self.device is not None:
            return self.device.readyForFluoScan
        else:
            return None

    def readyForManualTransfer(self):
        if self.device is not None:
            return self.device.readyForManualTransfert
        else:
            return None

    def readyForTransfer(self):
        if self.device is not None:
            return self.device.readyForTransfert
        else:
            return None

    def readyForVisuSample(self):
        if self.device is not None:
            return self.device.readyForVisuSample
        else:
            return None

    def gotoPhase(self, phase):
        logging.debug("PX1environment.gotoPhase %s" % phase)
        cmd = self.cmds.get(phase, None)
        if cmd is not None:
            logging.debug("PX1environment.gotoPhase state %s" % self.readState())
            cmd()
        else:
            return None

    def setPhase(self, phase, timeout=120):
        self.gotoPhase(phase)
        self.waitPhase(phase, timeout)

    def readPhase(self):
        if self.device is not None:
            phasename = self.device.currentPhase
            return EnvironmentPhase.phase(phasename)
        else:
            return None

    def get_current_phase(self):
        return self.device.currentPhase

    getCurrentPhase = get_current_phase

    def getPhase(self):
        if self.device is not None:
            phasename = self.device.currentPhase
            return phasename
        else:
            return None

    def waitPhase(self, phase, timeout=None):
        if self.device is None:
            return
        logging.debug("PX1environment: start waitPhase")
        _debut = time.time()
        n = 0
        with gevent.Timeout(
            timeout, Exception("Timeout waiting for environment phase")
        ):
            waiting = True
            while waiting:
                n += 1
                _phaseread = self.readPhase()
                if phase == _phaseread:
                    waiting = False
                gevent.sleep(0.05)
        logging.debug(
            "PX1environment: end waitPhase in %.1f sec N= %d"
            % ((time.time() - _debut), n)
        )

    #
    # ------- end phase handling

    def gotoCentringPhase(self):
        if not self.readyForCentring() or self.getPhase() != "CENTRING":
            self.get_command_object("GoToCentringPhase")()
            time.sleep(0.1)

    def gotoCollectPhase(self):
        if not self.readyForCollect() or self.getPhase() != "COLLECT":
            self.get_command_object("GoToCollectPhase")()
            time.sleep(0.1)

    def gotoLoadingPhase(self):
        if not self.readyForTransfer():
            self.get_command_object("GoToTransfertPhase")()
            time.sleep(0.1)

    def gotoManualLoadingPhase(self):
        if not self.readyForTransfer():
            self.get_command_object("GoToManualTransfertPhase")()
            time.sleep(0.1)

    def gotoSampleViewPhase(self):
        if not self.readyForVisuSample():
            self.get_command_object("GoToVisuSamplePhase")()
            time.sleep(0.1)

    def gotoFluoScanPhase(self):
        if not self.readyForFluoScan():
            self.get_command_object("GoToFluoScanPhase")()
            time.sleep(0.1)

    def setAuthorizationFlag(self, value):
        # make here the logic with eventually other permits (like hardware permit)
        if value != self.auth:
            self.log.debug(
                "PX1Environment. received authorization from cryotong:  %s" % value
            )
            self.auth = value
            self.emit("operationPermitted", value)

    def getUsingCapillary(self):
        if self.usingCapillaryChannel is not None:
            return self.usingCapillaryChannel.get_value()

    def setUsingCapillary(self, value):
        self.capillary_value = value
        gevent.spawn(self._setUsingCapillary)

    @task
    def _setUsingCapillary(self):
        if self.usingCapillaryChannel is not None:
            self.usingCapillaryChannel.set_value(self.capillary_value)

    def getBeamstopPosition(self):
        if self.beamstopPositionChannel is not None:
            return self.beamstopPositionChannel.get_value()

    def setBeamstopPosition(self, value):
        self.beamstop_position = value
        gevent.spawn(self._setBeamstopPosition)

    @task
    def _setBeamstopPosition(self):
        if self.beamstopPositionChannel is not None:
            self.beamstopPositionChannel.set_value(self.beamstop_position)


def test_hwo(hwo):
    t0 = time.time()
    print("PX1 Environment (state) ", hwo.get_state())
    print("               phase is ", hwo.getCurrentPhase())
    print("        beamstop pos is ", hwo.getBeamstopPosition())

    # if not env.readyForTransfer():
    #    print "Going to transfer phase"
    #    env.setPhase(EnvironmentPhase.TRANSFER)
    #    print time.time() - t0
    # print "done"
    # env.waitPhase(EnvironmentPhase.TRANSFER)



#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""MachineInfo using Tango channels.

Example xml_ configuration:

.. code-block:: xml

 <object class="TangoMachineInfo">
   <tangoname>ab/cd/ef</tangoname>
   <parameters>["current", "lifetime", "message", "refill_countdown"]</parameters>
   <channel type="tango" name="current" polling="2000">SR_Current</channel>
   <channel type="tango" name="lifetime" polling="2000">SR_Lifetime</channel>
   <channel type="tango" name="refill_countdown" polling="2000">SR_Refill_Countdown</channel>
   <channel type="tango" name="message" polling="8000">SR_Operator_Mesg</channel>
 </object>
"""

from mxcubecore.HardwareObjects.abstract.AbstractMachineInfo import AbstractMachineInfo

__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class TangoMachineInfo(AbstractMachineInfo):
    """MachineInfo using Tango channels."""

    def init(self):
        """We assume that at least current is defined"""
        super().init()
        # we only consider the attributes defined in the parameters property
        for name in self._mach_info_keys:
            if hasattr(self.get_channel_object(name), "get_value"):
                setattr(self, f"{name}", self.get_channel_object(name))
            else:
                # remove the attributes we cannot read
                self._mach_info_keys.pop(name)

        # we only want to update when the current changes.
        self.current.connect_signal("update", self._update_value)
        self.update_state(self.STATES.READY)

    def _check_attributes(self, attr_list=None):
        """Check if all the keys in the configuration file have
        implemented read method. Remove the undefined.
        """
        attr_list = attr_list or self._mach_info_keys

        for attr_key in attr_list:
            try:
                hasattr(self.get_channel_object(attr_key), "get_value")
            except AttributeError:
                attr_list.remove(attr_key)
        self._mach_info_keys = attr_list

    def _update_value(self, value):
        """Update all the attributes, not only the current."""
        self.update_value()

    def get_value(self) -> dict:
        """Read machine info summary as dictionary.

        Returns:
            Copy of the _mach_info_dict.
        """
        for name in self._mach_info_keys:
            try:
                self._mach_info_dict.update({name: getattr(self, name).get_value()})
            except Exception as err:
                self.log.exception(err)

        return self._mach_info_dict.copy()

    def get_current(self) -> float:
        """Read the ring current.

        Returns:
            Current [mA].
        """
        return self.current.get_value()



"""
Contains the classes

* ShapeHistory
* Shape
* Point
* Line
* CanvasGrid

ShapeHistory keeps track of the current shapes the user has created. The
shapes handled are any that inherits the Shape base class. There are currently
two shapes implemented Point and Line.

Point is the graphical representation of a centred position. A point can be
stored and managed by the ShapeHistory. the Line object represents a line
between two Point objects.

"""

import os

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.model import queue_model_objects

SELECTED_COLOR = "green"
NORMAL_COLOR = "yellow"


class ShapeHistoryMockup(HardwareObject):
    """
    Keeps track of the current shapes the user has created. The
    shapes handled are any that inherits the Shape base class.
    """

    def __init__(self, name):
        HardwareObject.__init__(self, name)
        self._drawing = None
        self.shapes = {}
        self.selected_shapes = {}
        self.point_index = 0

    def set_drawing(self, drawing):
        """
        Sets the drawing the Shape objects that are managed are drawn on.

        :param drawing: The drawing that the shapes are drawn on.
        :type drawing: QubDrawing (used by Qub)

        :returns: None
        """
        if self._drawing:
            self.log.info(
                "Setting previous drawing:" + str(self._drawing) + " to " + str(drawing)
            )

        self._drawing = drawing

    def get_drawing(self):
        """
        :returns: Returns the drawing of the shapes.
        :rtype: QubDrawing
        """
        return self._drawing

    def get_snapshot(self, qub_objects):
        """
        Get a snapshot of the video stream and overlay the objects
        in qub_objects.

        :param qub_objects: The QCanvas object to add on top of the video.
        :type qub_objects: QCanvas

        :returns: The snapshot
        :rtype: QImage
        """
        cwd = os.getcwd()
        path = os.path.join(cwd, "./test/HardwareObjectsMockup.xml/")
        qimg_path = os.path.join(path, "mxcube_sample_snapshot.jpeg")
        qimg = open(qimg_path, "rb").read()

        return qimg

    def get_qimage(self, image, canvas, zoom=1):
        """
        Gets the QImage from a parent widget.

        :param image: The QWidget that contains the image to extract
        :type image: QWidget

        :param canvas: The QCanvas object to add as overlay
        :type canvas: QCanvas

        :param zoom: Zoom level
        :type zoom: int.

        :returns: The QImage contained in the parent widget.
        :rtype: QImage
        """
        cwd = os.getcwd()
        path = os.path.join(cwd, "./test/HardwareObjectsMockup.xml/")
        img_path = os.path.join(path, "mxcube_sample_snapshot.jpeg")
        img = open(img_path, "rb").read()

        return img

    def get_shapes(self):
        """
        :returns: All the shapes currently handled.
        """
        return self.shapes.values()

    def get_points(self):
        """
        :returns: All points currently handled
        """
        current_points = []

        for shape in self.get_shapes():
            if isinstance(shape, Point):
                current_points.append(shape)

        return current_points

    def add_shape(self, shape):
        """
        Adds the shape <shape> to the list of handled objects.

        :param shape: Shape to add.
        :type shape: Shape object.

        """
        self.shapes[shape] = shape
        if isinstance(shape, Point):
            shape.set_index(self.get_available_point_index())

        self.get_drawing_event_handler().de_select_all()
        self.get_drawing_event_handler().set_selected(shape, True, call_cb=True)

    def get_available_point_index(self):
        self.point_index += 1
        return self.point_index

    def _delete_shape(self, shape):
        shape.unhighlight()

        if shape in self.selected_shapes:
            del self.selected_shapes[shape]

            if callable(self._drawing_event.selection_cb):
                self._drawing_event.selection_cb(self.selected_shapes.values())

        if shape is self._drawing_event.current_shape:
            self._drawing_event.current_shape = None

        if callable(self._drawing_event.deletion_cb):
            self._drawing_event.deletion_cb(shape)

    def delete_shape(self, shape):
        """
        Removes the shape <shape> from the list of handled shapes.

        :param shape: The shape to remove
        :type shape: Shape object.
        """
        related_points = []

        # If a point remove related line first
        if isinstance(shape, Point):
            for s in self.get_shapes():
                if isinstance(s, Line):
                    for s_qub_obj in s.get_qub_objects():
                        if s_qub_obj in shape.get_qub_objects():
                            self._delete_shape(s)
                            related_points.append(s)
                            break

        self._delete_shape(shape)
        del self.shapes[shape]

        # Delete the related shapes after self._delete_shapes so that
        # related object still exists when calling delete call back.
        for point in related_points:
            del self.shapes[point]

    def move_shape(self, shape, new_positions):
        """
        Moves the shape <shape> to the position <new_position>

        :param shape: The shape to move
        :type shape: Shape

        :param new_position: A tuple (X, Y)
        :type new_position: <int, int>
        """
        self.shapes[shape].move(new_positions)

    def clear_all(self):
        """
        Clear the shape history, remove all contents.
        """
        return

    def de_select_all(self):
        return

    def select_shape_with_cpos(self, cpos):
        return


class Shape(object):
    """
    Base class for shapes.
    """

    def __init__(self):
        object.__init__(self)
        self._drawing = None

    def get_drawing(self):
        """
        :returns: The drawing on which the shape is drawn.
        :rtype: QDrawing
        """
        return self._drawing

    def draw(self):
        """
        Draws the shape on its drawing.
        """
        pass

    def get_centred_positions(self):
        """
        :returns: The centred position(s) associated with the shape.
        :rtype: List of CentredPosition objects.
        """
        pass

    def hide(self):
        """
        Hides the shape.
        """
        pass

    def show(self):
        """
        Shows the shape.
        """
        pass

    def update_position(self):
        pass

    def move(self, new_positions):
        """
        Moves the shape to the position <new_position>
        """
        pass

    def highlight(self):
        """
        Highlights the shape
        """
        pass

    def unhighlight(self):
        """
        Removes highlighting.
        """
        pass

    def get_hit(self, x, y):
        """
        :returns: True if the shape was hit by the mouse.
        """
        pass

    def get_qub_objects(self):
        """
        :returns: A list of qub objects.
        """
        pass


class Line(Shape):
    def __init__(self, drawing, start_qub_p, end_qub_p, start_cpos, end_cpos):
        object.__init__(self)

        self._drawing = drawing
        self.start_qub_p = start_qub_p
        self.end_qub_p = end_qub_p
        self.start_cpos = start_cpos
        self.end_cpos = end_cpos
        self.qub_line = None

    def get_centred_positions(self):
        return [self.start_cpos, self.end_cpos]

    def hide(self):
        self.qub_line.hide()

    def show(self):
        self.qub_line.show()

    def update_position(self):
        self.qub_line.moveFirstPoint(self.start_qub_p._x, self.start_qub_p._y)
        self.qub_line.moveSecondPoint(self.start_qub_p._x, self.start_qub_p._y)

    def move(self, new_positions):
        self.qub_line.moveFirstPoint(new_positions[0][0], new_positions[0][1])
        self.qub_line.moveSecondPoint(new_positions[1][0], new_positions[1][1])

    def highlight(self):
        return

    def unhighlight(self):
        return

    def get_hit(self, x, y):
        return None
        # return self.qub_line.getModifyClass(x, y)

    def get_qub_objects(self):
        return [self.start_qub_p, self.end_qub_p, self.qub_line]

    def get_points_index(self):
        if self.start_cpos and self.end_cpos:
            return (self.start_cpos.get_index(), self.end_cpos.get_index())


class Point(Shape):
    def __init__(self, drawing, centred_position, screen_pos):
        Shape.__init__(self)

        self.qub_point = None

        if centred_position is None:
            self.centred_position = queue_model_objects.CentredPosition()
            self.centred_position.centring_method = False
        else:
            self.centred_position = centred_position
        self.screen_pos = screen_pos
        self.point_index = None
        self._drawing = drawing

        self.qub_point = self.draw(screen_pos)

    def set_index(self, index):
        self.point_index = index

    def get_index(self):
        return self.point_index

    def get_qub_point(self):
        return self.qub_point

    def get_centred_positions(self):
        return [self.centred_position]

    def draw(self, screen_pos):
        """
        Draws a qub point in the sample video.
        """
        qub_point = None

        return qub_point

    def show(self):
        return

    def hide(self):
        return

    def move(self, new_positions):
        return

    def highlight(self):
        return

    def unhighlight(self):
        return



import logging
import time

import gevent

from mxcubecore.HardwareObjects.Cats90 import (
    TOOL_SPINE,
    Cats90,
    SampleChangerState,
)

TIMEOUT = 3


class ALBACats(Cats90):
    """
    Main class used @ ALBA to integrate the CATS-IRELEC sample changer.
    """

    def __init__(self, *args):
        Cats90.__init__(self, *args)

    def init(self):
        Cats90.init(self)

        self.shifts_channel = self.get_channel_object("shifts")
        self.phase_channel = self.get_channel_object("phase")

        self.go_transfer_cmd = self.get_command_object("go_transfer")
        self.go_sampleview_cmd = self.get_command_object("go_sampleview")
        self.super_abort_cmd = self.get_command_object("super_abort")
        self.super_state_channel = self.get_channel_object("super_state")

        self.auto_prepare_diff = self.get_property("auto_prepare_diff")
        self.detdist_position_channel = self.get_channel_object("detdist_position")
        self.detdist_saved = None

        self._cmdLoadHT = self.get_command_object("_cmdLoadHT")
        self._cmdChainedLoadHT = self.get_command_object("_cmdChainedLoadHT")
        self._cmdUnloadHT = self.get_command_object("_cmdUnloadHT")

        self._chnPathSafe = self.get_channel_object("_chnPathSafe")

        if self._chnPathRunning is not None:
            self._chnPathRunning.connect_signal("update", self._update_running_state)

        if self._chnPowered is not None:
            self._chnPowered.connect_signal("update", self._update_powered_state)

    def is_ready(self):
        """
        Returns a boolean value indicating is the sample changer is ready for operation.

        @return: boolean
        """
        return (
            self.state == SampleChangerState.Ready
            or self.state == SampleChangerState.Loaded
            or self.state == SampleChangerState.Charging
            or self.state == SampleChangerState.StandBy
            or self.state == SampleChangerState.Disabled
        )

    def diff_send_transfer(self):
        """
        Checks if beamline supervisor is in TRANSFER phase (i.e. sample changer in TRANSFER phase too).
        If is not the case, It sends the sample changer to TRANSFER phase.
        Returns a boolean value indication if the sample changer is in TRANSFER phase.

        @return: boolean
        """
        if self.read_super_phase().upper() == "TRANSFER":
            logging.getLogger("user_level_log").error(
                "Supervisor is already in transfer phase"
            )
            return True

        self.go_transfer_cmd()
        ret = self._wait_phase_done("TRANSFER")
        return ret

    def diff_send_sampleview(self):
        """
        Checks if beamline supervisor is in SAMPLE phase (i.e. sample changer in SAMPLE phase too).
        If is not the case, It sends the sample changer to SAMPLE phase.
        Returns a boolean value indication if the sample changer is in SAMPLE phase.

        @return: boolean
        """
        if self.read_super_phase().upper() == "SAMPLE":
            logging.getLogger("user_level_log").error(
                "Supervisor is already in sample view phase"
            )
            return True

        t0 = time.time()
        while True:
            state = str(self.super_state_channel.get_value())
            if str(state) == "ON":
                break

            if (time.time() - t0) > TIMEOUT:
                logging.getLogger("user_level_log").error(
                    "Supervisor timeout waiting for ON state. Returning"
                )
                return False

            time.sleep(0.1)

        self.go_sampleview_cmd()
        ret = self._wait_phase_done("SAMPLE")
        return ret

    def _wait_super_ready(self):
        while True:
            state = str(self.super_state_channel.get_value())
            if state == "ON":
                logging.getLogger("user_level_log").error(
                    "Supervisor is in ON state. Returning"
                )
                break

    def _wait_phase_done(self, final_phase):
        """
        Method to wait a phase change. When supervisor reaches the final phase, the diffracometer
        returns True.

        @final_phase: target phase
        @return: boolean
        """
        while True:
            state = str(self.super_state_channel.get_value())
            if state == "ON":
                logging.getLogger("user_level_log").error(
                    "Supervisor is in ON state. Returning"
                )
                break
            elif str(state) != "MOVING":
                logging.getLogger("user_level_log").error(
                    "Supervisor is in a funny state %s" % str(state)
                )
                return False

            self.log.debug("Supervisor waiting to finish phase change")
            time.sleep(0.2)

        time.sleep(0.1)

        if self.read_super_phase().upper() != final_phase:
            logging.getLogger("user_level_log").error(
                "Supervisor is not yet in %s phase. Aborting load" % final_phase
            )
            return False
        else:
            return True

    def save_detdist_position(self):
        self.detdist_saved = self.detdist_position_channel.get_value()
        logging.getLogger("user_level_log").error(
            "Saving current det.distance (%s)" % self.detdist_saved
        )

    def restore_detdist_position(self):
        if abs(self.detdist_saved - self.detdist_position_channel.get_value()) >= 0.1:
            logging.getLogger("user_level_log").error(
                "Restoring det.distance to %s" % self.detdist_saved
            )
            self.detdist_position_channel.set_value(self.detdist_saved)
            time.sleep(0.4)
            self._wait_super_ready()

    def read_super_phase(self):
        """
        Returns supervisor phase (CurrentPhase attribute from Beamline Supervisor Tango DS)

        @return: str
        """
        return self.phase_channel.get_value()

    def load(self, sample=None, wait=False, wash=False):
        """
        Loads a sample. Overrides to include ht basket.

        @sample: sample to load.
        @wait:
        @wash: wash dring the load operation.
        @return:
        """

        self.log.debug("Loading sample %s / type(%s)" % (sample, type(sample)))

        ret, msg = self._check_coherence()
        if not ret:
            raise Exception(msg)

        sample_ht = self.is_ht_sample(sample)

        if not sample_ht:
            sample = self._resolve_component(sample)
            self.assert_not_charging()
            use_ht = False
        else:
            sample = sample_ht
            use_ht = True

        if self.has_loaded_sample():
            if (wash is False) and self.get_loaded_sample() == sample:
                raise Exception(
                    "The sample " + sample.get_address() + " is already loaded"
                )
            else:
                # Unload first / do a chained load
                pass

        return self._execute_task(
            SampleChangerState.Loading, wait, self._do_load, sample, None, use_ht
        )

    def unload(self, sample_slot=None, wait=False):
        """
        Unload the sample. If sample_slot=None, unloads to the same slot the sample was loaded from.

        @sample_slot:
        @wait:
        @return:
        """
        sample_slot = self._resolve_component(sample_slot)

        self.assert_not_charging()

        # In case we have manually mounted we can command an unmount
        if not self.has_loaded_sample():
            raise Exception("No sample is loaded")

        return self._execute_task(
            SampleChangerState.Unloading, wait, self._do_unload, sample_slot
        )

    # TODO: this overrides identical method from Cats90
    def is_powered(self):
        return self._chnPowered.get_value()

    # TODO: this overrides identical method from Cats90
    def is_path_running(self):
        return self._chnPathRunning.get_value()

    # TODO: this overrides method from AbstractSampleChanger
    # def has_loaded_sample(self):  # not used.  to use it remove _
    #   return self._chnSampleIsDetected.get_value()

    def _update_running_state(self, value):
        """
        Emits signal with new Running State

        @value: New running state
        """
        self.emit("runningStateChanged", (value,))

    def _update_powered_state(self, value):
        """
        Emits signal with new Powered State

        @value: New powered state
        """
        self.emit("powerStateChanged", (value,))

    def _do_load(self, sample=None, shifts=None, use_ht=False):
        """
        Loads a sample on the diffractometer. Performs a simple put operation if the diffractometer is empty, and
        a sample exchange (unmount of old + mount of new sample) if a sample is already mounted on the diffractometer.
        Overrides Cats90 method.

        @sample: sample to load.
        @shifts: mounting point offsets.
        @use_ht: mount a sample from hot tool.
        """
        if not self._chnPowered.get_value():
            self._cmdPowerOn()  # try switching power on

        current_tool = self.get_current_tool()

        self.save_detdist_position()
        ret = self.diff_send_transfer()

        if ret is False:
            logging.getLogger("user_level_log").error(
                "Supervisor cmd transfer phase returned an error."
            )
            self._update_state()  # remove transient states like Loading. Reflect hardware state
            raise Exception(
                "CATS cannot get to transfer phase. Aborting sample changer operation."
            )

        gevent.sleep(3)
        if not self._chnPowered.get_value():
            raise Exception(
                "CATS power is not enabled. Please switch on arm power before transferring samples."
            )

        # obtain mounting offsets from diffr
        shifts = self._get_shifts()

        if shifts is None:
            xshift, yshift, zshift = ["0", "0", "0"]
        else:
            xshift, yshift, zshift = map(str, shifts)

        # get sample selection
        selected = self.get_selected_sample()

        self.log.debug(
            "  ==========CATS=== selected sample is %s (prev %s)"
            % (str(selected), str(sample))
        )

        if not use_ht:
            if sample is not None:
                if sample != selected:
                    self._do_select(sample)
                    selected = self.get_selected_sample()
            else:
                if selected is not None:
                    sample = selected
                else:
                    raise Exception("No sample selected")
        else:
            selected = None

        # some cancel cases
        if (
            not use_ht
            and self.has_loaded_sample()
            and selected == self.get_loaded_sample()
        ):
            self._update_state()  # remove transient states like Loading. Reflect hardware state
            raise Exception(
                "The sample "
                + str(self.get_loaded_sample().get_address())
                + " is already loaded"
            )

        if not self.has_loaded_sample() and self.cats_sample_on_diffr() == 1:
            self.log.warning("  ==========CATS=== sample on diffr, loading aborted")
            self._update_state()  # remove transient states like Loading. Reflect hardware state
            raise Exception(
                "The sample "
                + str(self.get_loaded_sample().get_address())
                + " is already loaded"
            )
            return

        if self.cats_sample_on_diffr() == -1:
            self._update_state()  # remove transient states like Loading. Reflect hardware state
            raise Exception(
                "Conflicting info between diffractometer and on-magnet detection. Consider 'Clear'"
            )
            return

        # end some cancel cases

        # if load_ht
        loaded_ht = self.is_loaded_ht()

        #
        # Loading HT sample
        #
        if use_ht:  # loading HT sample
            if loaded_ht == -1:  # has loaded but it is not HT
                # first unmount (non HT)
                self.log.warning(
                    "  ==========CATS=== mix load/unload dewar vs HT (NOT IMPLEMENTED YET)"
                )
                return

            argin = ["2", str(sample), "0", "0", xshift, yshift, zshift]
            self.log.warning("  ==========CATS=== about to load HT. %s" % str(argin))
            if loaded_ht == 1:  # has ht loaded
                cmd_ok = self._execute_server_task(
                    self._cmdChainedLoadHT, argin, waitsafe=True
                )
            else:
                cmd_ok = self._execute_server_task(
                    self._cmdLoadHT, argin, waitsafe=False
                )

        #
        # Loading non HT sample
        #
        else:
            if loaded_ht == 1:  # has an HT sample mounted
                # first unmount HT
                self.log.warning(
                    "  ==========CATS=== mix load/unload dewar vs HT (NOT IMPLEMENTED YET)"
                )
                return

            # calculate CATS specific lid/sample number
            # lid = ((selected.get_basket_no() - 1) / 3) + 1
            # sample = (((selected.get_basket_no() - 1) % 3) * 10) + selected.get_vial_no()

            basketno = selected.get_basket_no()
            sampleno = selected.get_vial_no()

            lid, sample = self.basketsample_to_lidsample(basketno, sampleno)
            tool = self.tool_for_basket(basketno)
            stype = self.get_cassette_type(basketno)

            if tool != current_tool:
                self.log.warning(
                    "  ==========CATS=== changing tool from %s to %s"
                    % (current_tool, tool)
                )
                changing_tool = True
            else:
                changing_tool = False

            # we should now check basket type on diffr to see if tool is different...
            # then decide what to do

            if shifts is None:
                xshift, yshift, zshift = ["0", "0", "0"]
            else:
                xshift, yshift, zshift = map(str, shifts)

            # prepare argin values
            argin = [
                str(tool),
                str(lid),
                str(sample),
                str(stype),
                "0",
                xshift,
                yshift,
                zshift,
            ]

            if tool == 2:
                read_barcode = (
                    self.read_datamatrix and self._cmdChainedLoadBarcode is not None
                )
            else:
                if self.read_datamatrix:
                    self.log.warning(
                        "  ==========CATS=== reading barcode only possible with spine pucks"
                    )
                read_barcode = False

            if loaded_ht == -1:  # has a loaded but it is not an HT
                if changing_tool:
                    raise Exception(
                        "This operation requires a tool change. You should unload sample first"
                    )

                if read_barcode:
                    self.log.warning(
                        "  ==========CATS=== chained load sample (barcode), sending to cats:  %s"
                        % argin
                    )
                    cmd_ok = self._execute_server_task(
                        self._cmdChainedLoadBarcode, argin, waitsafe=True
                    )
                else:
                    self.log.warning(
                        "  ==========CATS=== chained load sample, sending to cats:  %s"
                        % argin
                    )
                    cmd_ok = self._execute_server_task(
                        self._cmdChainedLoad, argin, waitsafe=True
                    )
            elif loaded_ht == 0:
                if read_barcode:
                    self.log.warning(
                        "  ==========CATS=== load sample (barcode), sending to cats:  %s"
                        % argin
                    )
                    cmd_ok = self._execute_server_task(
                        self._cmdLoadBarcode, argin, waitsafe=True
                    )
                else:
                    self.log.warning(
                        "  ==========CATS=== load sample, sending to cats:  %s" % argin
                    )
                    cmd_ok = self._execute_server_task(
                        self._cmdLoad, argin, waitsafe=True
                    )

        if not cmd_ok:
            self.log.info("  LOAD Command failed on device server")
        elif self.auto_prepare_diff and not changing_tool:
            self.log.info(
                "  AUTO_PREPARE_DIFF (On) sample changer is in safe state... preparing diff now"
            )
            # ret = self.diff_send_sampleview()
            self.go_sampleview_cmd()
            self.log.info("     restoring detector distance")
            self.restore_detdist_position()
            self._wait_phase_done("SAMPLE")
        else:
            self.log.info(
                "  AUTO_PREPARE_DIFF (Off) sample loading done / or changing tool (%s)"
                % changing_tool
            )

        # load commands are executed until path is safe. Then we have to wait for
        # path to be finished
        self._wait_device_ready()

    def _do_unload(self, sample_slot=None, shifts=None):
        """
        Unloads a sample from the diffractometer.
        Overrides Cats90 method.

        @sample_slot:
        @shifts: mounting position
        """
        if not self._chnPowered.get_value():
            self._cmdPowerOn()  # try switching power on

        ret = self.diff_send_transfer()

        if ret is False:
            logging.getLogger("user_level_log").error(
                "Supervisor cmd transfer phase returned an error."
            )
            return

        shifts = self._get_shifts()

        if sample_slot is not None:
            self._do_select(sample_slot)

        loaded_ht = self.is_loaded_ht()

        if shifts is None:
            xshift, yshift, zshift = ["0", "0", "0"]
        else:
            xshift, yshift, zshift = map(str, shifts)

        loaded_lid = self._chnLidLoadedSample.get_value()
        loaded_num = self._chnNumLoadedSample.get_value()

        if loaded_lid == -1:
            self.log.warning(
                "  ==========CATS=== unload sample, no sample mounted detected"
            )
            return

        loaded_basket, loaded_sample = self.lidsample_to_basketsample(
            loaded_lid, loaded_num
        )

        tool = self.tool_for_basket(loaded_basket)

        argin = [str(tool), "0", xshift, yshift, zshift]

        self.log.warning(
            "  ==========CATS=== unload sample, sending to cats:  %s" % argin
        )
        if loaded_ht == 1:
            cmd_ret = self._execute_server_task(self._cmdUnloadHT, argin)
        else:
            cmd_ret = self._execute_server_task(self._cmdUnload, argin)

    def _do_abort(self):
        """
        Aborts a running trajectory on the sample changer.

        :returns: None
        :rtype: None
        """
        if self.super_abort_cmd is not None:
            self.super_abort_cmd()  # stops super
        self._cmdAbort()
        self._update_state()  # remove software flags like Loading.. reflects current hardware state

    def _check_coherence(self):
        detected = self._chnSampleIsDetected.get_value()
        loaded_lid = self._chnLidLoadedSample.get_value()
        loaded_num = self._chnNumLoadedSample.get_value()

        if -1 in [loaded_lid, loaded_num] and detected:
            return False, "Sample detected on Diffract. but there is no info about it"

        return True, ""

    def _get_shifts(self):
        """
        Get the mounting position from the Diffractometer DS.

        @return: 3-tuple
        """
        if self.shifts_channel is not None:
            shifts = self.shifts_channel.get_value()
        else:
            shifts = None
        return shifts

    # def path_running(self):
    # """
    # Overrides Cats90 method.
    #
    # @return:
    # """
    # return (self._chnPathSafe.get_value() is not True)

    # TODO: fix return type
    def is_ht_sample(self, address):
        """
        Returns is sample address belongs to hot tool basket.

        @address: sample address
        @return: int or boolean
        """
        basket, sample = address.split(":")
        try:
            if int(basket) >= 100:
                return int(sample)
            else:
                return False
        except Exception:
            return False

    def tool_for_basket(self, basketno):
        """
        Returns the tool corresponding to the basket.

        @basketno: basket number
        @return: int
        """
        if basketno == 100:
            return TOOL_SPINE

        return Cats90.tool_for_basket(self, basketno)

    def is_loaded_ht(self):
        """
         1 : has loaded ht
         0 : nothing loaded
        -1 : loaded but not ht
        """
        sample_lid = self._chnLidLoadedSample.get_value()

        if self.has_loaded_sample():
            if sample_lid == 100:
                return 1
            else:
                return -1
        else:
            return 0


def test_hwo(hwo):
    hwo._update_cats_contents()
    print(" Is path running? ", hwo.is_path_running())
    print(" Loading shifts:  ", hwo._get_shifts())
    print(" Sample on diffr :  ", hwo.cats_sample_on_diffr())
    print(" Baskets :  ", hwo.basket_presence)
    print(" Baskets :  ", hwo.get_basket_list())
    if hwo.has_loaded_sample():
        print(" Loaded is: ", hwo.get_loaded_sample().get_coords())
    print(" Is mounted sample: ", hwo.is_mounted_sample((1, 1)))


if __name__ == "__main__":
    test()



# encoding: utf-8
"""Abstract beamline interface message classes

License:

This file is part of MXCuBE.

MXCuBE is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MXCuBE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with MXCuBE. If not, see <https://www.gnu.org/licenses/>.
"""

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

import json
import uuid
from collections import (
    OrderedDict,
    namedtuple,
)

from mxcubecore.model import crystal_symmetry

__copyright__ = """ Copyright © 2016 - 2019 by Global Phasing Ltd. """
__license__ = "LGPLv3+"
__author__ = "Rasmus H Fogh"


# Enumerations

MESSAGE_INTENTS = {"DOCUMENT", "COMMAND", "EVENT"}

INDEXING_FORMATS = ("IDXREF",)

ABSORPTION_EDGES = ("K", "LI", "LII", "LIII", "MI", "MII", "MIII", "MIV", "MV")

CHEMICAL_ELEMENTS = OrderedDict(
    (
        ("H", "hydrogen"),
        ("HE", "helium"),
        ("LI", "lithium"),
        ("BE", "beryllium"),
        ("B", "boron"),
        ("C", "carbon"),
        ("N", "nitrogen"),
        ("O", "oxygen"),
        ("F", "fluorine"),
        ("NE", "neon"),
        ("NA", "sodium"),
        ("MG", "magnesium"),
        ("AL", "aluminium"),
        ("SI", "silicon"),
        ("P", "phosphorus"),
        ("S", "sulfur"),
        ("CL", "chlorine"),
        ("AR", "argon"),
        ("K", "potassium"),
        ("CA", "calcium"),
        ("SC", "scandium"),
        ("TI", "titanium"),
        ("V", "vanadium"),
        ("CR", "chromium"),
        ("MN", "manganese"),
        ("FE", "iron"),
        ("CO", "cobalt"),
        ("NI", "nickel"),
        ("CU", "copper"),
        ("ZN", "zinc"),
        ("GA", "gallium"),
        ("GE", "germanium"),
        ("AS", "arsenic"),
        ("SE", "selenium"),
        ("BR", "bromine"),
        ("KR", "krypton"),
        ("RB", "rubidium"),
        ("SR", "strontium"),
        ("Y", "yttrium"),
        ("ZR", "zirconium"),
        ("NB", "niobium"),
        ("MO", "molybdenum"),
        ("TC", "technetium"),
        ("RU", "ruthenium"),
        ("RH", "rhodium"),
        ("PD", "palladium"),
        ("AG", "silver"),
        ("CD", "cadmium"),
        ("IN", "indium"),
        ("SN", "tin"),
        ("SB", "antimony"),
        ("TE", "tellurium"),
        ("I", "iodine"),
        ("XE", "xenon"),
        ("CS", "caesium"),
        ("BA", "barium"),
        ("LA", "lanthanum"),
        ("CE", "cerium"),
        ("PR", "praseodymium"),
        ("ND", "neodymium"),
        ("PM", "promethium"),
        ("SM", "samarium"),
        ("EU", "europium"),
        ("GD", "gadolinium"),
        ("TB", "terbium"),
        ("DY", "dysprosium"),
        ("HO", "holmium"),
        ("ER", "erbium"),
        ("TM", "thulium"),
        ("YB", "ytterbium"),
        ("LU", "lutetium"),
        ("HF", "hafnium"),
        ("TA", "tantalum"),
        ("W", "tungsten"),
        ("RE", "rhenium"),
        ("OS", "osmium"),
        ("IR", "iridium"),
        ("PT", "platinum"),
        ("AU", "gold"),
        ("HG", "mercury"),
        ("TL", "thallium"),
        ("PB", "lead"),
        ("BI", "bismuth"),
        ("PO", "polonium"),
        ("AT", "astatine"),
        ("RN", "radon"),
        ("FR", "francium"),
        ("RA", "radium"),
        ("AC", "actinium"),
        ("TH", "thorium"),
        ("PA", "protactinium"),
        ("U", "uranium"),
        ("NP", "neptunium"),
        ("PU", "plutonium"),
        ("AM", "americium"),
        ("CM", "curium"),
        ("BK", "berkelium"),
        ("CF", "californium"),
        ("ES", "einsteinium"),
        ("FM", "fermium"),
        ("MD", "mendelevium"),
        ("NO", "nobelium"),
        ("LR", "lawrencium"),
        ("RF", "rutherfordium"),
        ("DB", "dubnium"),
        ("SG", "seaborgium"),
        ("BH", "bohrium"),
        ("HS", "hassium"),
        ("MT", "meitnerium"),
        ("DS", "darmstadtium"),
        ("RG", "roentgenium"),
        ("CN", "copernicium"),
        ("UUT", "ununtrium"),
        ("FL", "flerovium"),
        ("UUP", "ununpentium"),
        ("LV", "livermorium"),
    )
)

ParsedMessage = namedtuple(
    "ParsedMessage", ("message_type", "payload", "enactment_id", "correlation_id")
)


# Abstract classes


class MessageData(object):
    """Topmost superclass for all message data objects

    Later to add e.g. stringification"""


class Payload(MessageData):
    """Payload - top level message object"""

    INTENT = None

    def __init__(self):
        # This class is abstract
        intent = self.__class__.INTENT
        if intent not in MESSAGE_INTENTS:
            if intent is None:
                raise RuntimeError("Attempt to instantiate abstract class Payload")
            raise RuntimeError(
                "Programming error - "
                "Payload subclass %s intent %s must be one of: %s"
                % (self.__class__.__name__, intent, sorted(MESSAGE_INTENTS))
            )

    @property
    def intent(self):
        """Message intent - class-level property"""
        return self.__class__.INTENT


class IdentifiedElement(MessageData):
    """Object with persistent uuid"""

    def __init__(self, id_=None):
        # This class is abstract
        if self.__class__.__name__ == "IdentifiedElement":
            raise RuntimeError(
                "Attempt to instantiate abstract class IdentifiedElement"
            )

        self._id = None
        self.__set_id(id_)

    @property
    def id_(self):
        """Unique identifier (UUID) for IdentifiedElement.
        Defaults to new, time-based uuid"""
        return self._id

    def __set_id(self, value):
        """Setter for uuid - accessible only within this class"""
        if value is None:
            self._id = uuid.uuid1()
        elif isinstance(value, uuid.UUID):
            self._id = value
        else:
            raise TypeError("UUID input must be of type uuid.UUID")


# Sync with Java 4/5/2017
# Acknowledge???
# Images???
# OrientationMatrix. NB this is a JAva stub. Not needed for now
# ObtainPriorInformation
# PrepareForCentring
# ReadyForCentring

# Intent is now  DOCUMENT, COMMAND, EVENT # NBNB TODO
# (data, command, info ca.) I could skip them?

# Simple payloads


class RequestConfiguration(Payload):
    """Configuration request message"""

    INTENT = "COMMAND"


class ObtainPriorInformation(Payload):
    """Prior information request"""

    INTENT = "COMMAND"


class PrepareForCentring(Payload):
    """Prior information request"""

    INTENT = "COMMAND"


class ReadyForCentring(Payload):
    """Prior information request"""

    INTENT = "DOCUMENT"


class SubprocessStopped(Payload):
    """Subprocess Stopped request message"""

    INTENT = "EVENT"


class ConfigurationData(Payload):
    """Configuration Data message"""

    INTENT = "DOCUMENT"

    # NB coded as mandatory, even if not explicitly non-null
    # (but raises MalformedUrlException) in Java.

    def __init__(self, location):
        super().__init__()
        self._location = location

    @property
    def location(self):
        """Url for directory containing configuration data.
        Generally an absolute file path."""
        return self._location


class SubprocessStarted(Payload):
    """Subprocess Started message"""

    INTENT = "EVENT"

    def __init__(self, name):
        super().__init__()
        self._name = name

    @property
    def name(self):
        """name of subprocess"""
        return self._name


class ChooseLattice(Payload):
    """Choose lattice instruction"""

    INTENT = "COMMAND"

    def __init__(
        self,
        indexingSolutions,
        indexingFormat="IDXREF",
        indexingHeader=None,
        priorCrystalClasses=(),
        priorSpaceGroup=None,
        priorSpaceGroupString=None,
        userProvidedCell=None,
    ):
        """

        Args:
            indexingSolutions (list(IndexingSolution):
            indexingFormat (str):
            indexingHeader str:
            priorCrystalClasses sequence(str):
            priorSpaceGroup int:
            priorSpaceGroupString str:
            userProvidedCell UnitCell:
        """
        super().__init__()

        self._indexingSolutions = indexingSolutions
        self._indexingFormat = indexingFormat
        self._indexingHeader = indexingHeader
        self._priorCrystalClasses = priorCrystalClasses or ()
        self._priorSpaceGroup = priorSpaceGroup
        self._priorSpaceGroupString = priorSpaceGroupString
        self._userProvidedCell = userProvidedCell

    @property
    def priorCrystalClasses(self):
        """tuple of crystal class names"""
        return self._priorCrystalClasses

    @property
    def priorSpaceGroup(self):
        """space group number"""
        return self._priorSpaceGroup

    @property
    def priorSpaceGroupString(self):
        """space group name"""
        return self._priorSpaceGroupString

    @property
    def indexingSolutions(self):
        """List of IndexingSolution"""
        return self._indexingSolutions

    @property
    def indexingFormat(self):
        """Indexing format"""
        return self._indexingFormat

    @property
    def indexingHeader(self):
        """Indexing header"""
        return self._indexingHeader

    @property
    def userProvidedCell(self):
        """User provided unit cell"""
        return self._userProvidedCell


class SelectedLattice(MessageData):
    """Lattice selected message"""

    INTENT = "DOCUMENT"

    def __init__(
        self,
        data_model,
        solution,
    ):
        self._solution = solution
        self._strategyDetectorSetting = data_model.detector_setting
        self._strategyWavelength = data_model.wavelengths[0]
        self._strategyControl = json.dumps(data_model.strategy_options, sort_keys=True)
        self._userCrystalClasses = data_model.crystal_classes
        sginfo = crystal_symmetry.SPACEGROUP_MAP.get(data_model.space_group)
        self._userSpaceGroup = sginfo.number if sginfo else None
        if data_model.reference_reflection_files:
            self._referenceReflectionFiles = list(data_model.reference_reflection_files)
        else:
            self._referenceReflectionFiles = []

    @property
    def solution(self):
        """Proposed solution"""
        return self._solution

    @property
    def strategyDetectorSetting(self):
        """Detector setting to use for strategy calculation and acquisition"""
        return self._strategyDetectorSetting

    @property
    def strategyWavelength(self):
        """Wavelength to use for strategy calculation and acquisition"""
        return self._strategyWavelength

    @property
    def strategyControl(self):
        """JSON string of command line options (*without* prefix)
        to use for stratcal wrapper call"""
        return self._strategyControl

    @property
    def userCrystalClasses(self):
        """Crystal classes given by user"""
        return self._userCrystalClasses

    @property
    def userSpaceGroup(self):
        """Space group (int) given by user"""
        return self._userSpaceGroup

    @property
    def referenceReflectionFiles(self):
        return list(self._referenceReflectionFiles)


class IndexingSolution(MessageData):
    """Indexing solution data"""

    def __init__(
        self,
        bravaisLattice,
        cell,
        isConsistent=None,
        latticeCharacter=None,
        qualityOfFit=None,
    ):
        """

        Args:
            bravaisLattice (string): One of the 14 Bravais lattices ('aP' etc.)
            cell (UnitCell):
            isConsistent (bool): Is solution consistent with know symmetry?
            latticeCharacter (int):  Integer 1-44
            qualityOfFit (float):
        """
        self._bravaisLattice = bravaisLattice
        self._cell = cell
        self._isConsistent = isConsistent
        self._latticeCharacter = latticeCharacter
        self._qualityOfFit = qualityOfFit

    @property
    def bravaisLattice(self):
        """One of the 14 Bravais lattices ('aP' etc.)"""
        return self._bravaisLattice

    @property
    def cell(self):
        """Unit ce;;"""
        return self._cell

    @property
    def isConsistent(self):
        """Is solution consistent with know symmetry?"""
        return self._isConsistent

    @property
    def latticeCharacter(self):
        """Integer 1-44"""
        return self._latticeCharacter

    @property
    def qualityOfFit(self):
        """"""
        return self._qualityOfFit


class CollectionDone(MessageData):
    """Collection Done message"""

    INTENT = "EVENT"

    def __init__(
        self,
        proposalId,
        status,
        procWithLatticeParams=False,
        imageRoot=None,
        scanIdMap=None,
        centrings=None,
    ):
        self._proposalId = proposalId
        self._imageRoot = imageRoot
        self._status = status
        self._procWithLatticeParams = procWithLatticeParams
        self._scanIdMap = scanIdMap
        self._centrings = centrings

    @property
    def proposalId(self):
        """uuid of collection proposal that has been executed."""
        return self._proposalId

    @property
    def imageRoot(self):
        """Url for directory containing images.
        Generally an absolute file path."""
        return self._imageRoot

    @property
    def status(self):
        """Integer status code for collection result"""
        return self._status

    @property
    def procWithLatticeParams(self):
        """Boolean, whether lattice parameters should be used for processing"""
        return self._procWithLatticeParams

    @property
    def scanIdMap(self):
        """Dict[str,str] scan.id_:GoniostatTranslation.id_"""
        return self._scanIdMap

    @property
    def centrings(self):
        """set(GoniostatTranslation)

        New GoniostatTranslations acquired during acquisition"""
        return self._centrings


# Complex payloads


class WorkflowDone(Payload):
    """End-of-workflow message"""

    INTENT = "EVENT"

    def __init__(self, issues=None):
        super().__init__()

        if self.__class__.__name__ == "WorkflowDone":
            raise RuntimeError("Attempt to instantiate abstract class WorkflowDone")

        if issues:
            ll0 = list(x for x in issues if not isinstance(x, Issue))
            if ll0:
                raise ValueError("issues parameter contains non-issues: %s" % ll0)
            else:
                self._issues = tuple(issues)


class WorkflowCompleted(WorkflowDone):
    pass


class WorkflowAborted(WorkflowDone):
    pass


class WorkflowFailed(WorkflowDone):
    def __init__(self, reason=None, issues=None):
        super(WorkflowFailed, self).__init__(issues=issues)
        self._reason = reason

    @property
    def reason(self):
        return self._reason


class BeamlineAbort(Payload):
    """Abort workflow from beamline"""

    INTENT = "COMMAND"


# Simple data objects


class AnomalousScatterer(MessageData):
    def __init__(self, element, edge):
        if element in CHEMICAL_ELEMENTS:
            self._element = element
        else:
            raise ValueError("Chemical element code %s not recognised" % element)

        if edge in ABSORPTION_EDGES:
            self._edge = edge
        else:
            raise ValueError("Absorption edge code %s not recognised" % edge)

    @property
    def element(self):
        return self._element

    @property
    def edge(self):
        return self._edge


class UnitCell(MessageData):
    """Unit cell data type"""

    def __init__(self, a, b, c, alpha, beta, gamma):
        self._lengths = (a, b, c)
        self._angles = (alpha, beta, gamma)

    @property
    def lengths(self):
        return self._lengths

    @property
    def angles(self):
        return self._angles

    @property
    def a(self):
        return self._lengths[0]

    @property
    def b(self):
        return self._lengths[1]

    @property
    def c(self):
        return self._lengths[2]

    @property
    def alpha(self):
        return self._angles[0]

    @property
    def beta(self):
        return self._angles[1]

    @property
    def gamma(self):
        return self._angles[2]


class Issue(IdentifiedElement):
    """Issue (status information returned with WorkflowDone messages)"""

    def __init__(self, component, message, code=None, id_=None):
        IdentifiedElement.__init__(self, id_)
        self._component = component
        self._message = message
        self._code = code

    @property
    def component(self):
        """Part of unique identifier of issue definition"""
        return self._component

    @property
    def code(self):
        """Part of unique identifier of issue definition"""
        return self._code

    @property
    def message(self):
        """Text providing specific details about this issue"""
        return self._message


class PhasingWavelength(IdentifiedElement):
    """Phasing Wavelength"""

    def __init__(self, wavelength, role=None, id_=None):
        IdentifiedElement.__init__(self, id_)
        self._role = role
        self._wavelength = wavelength

    @property
    def role(self):
        """Wavelength role"""
        return self._role

    @property
    def wavelength(self):
        """Wavelength setting for beam"""
        return self._wavelength

    @wavelength.setter
    def wavelength(self, value):
        if self._wavelength:
            raise TypeError("PhasingWavelength values cannot be re-set if non-zero")
        self._wavelength = value


class BeamSetting(IdentifiedElement):
    """Beam setting"""

    def __init__(self, wavelength, id_=None):
        IdentifiedElement.__init__(self, id_)
        self._wavelength = wavelength

    @property
    def wavelength(self):
        """Wavelength setting for beam"""
        return self._wavelength


class ScanExposure(IdentifiedElement):
    """Scan Exposure"""

    def __init__(self, time, transmission, id_=None):
        IdentifiedElement.__init__(self, id_)
        self._time = time
        self._transmission = transmission

    @property
    def transmission(self):
        """Scan exposure transmission"""
        return self._transmission

    @property
    def time(self):
        """Scan exposure transmission"""
        return self._time


class ScanWidth(IdentifiedElement):
    """Scan Width"""

    def __init__(self, imageWidth, numImages, id_=None):
        IdentifiedElement.__init__(self, id_)
        self._imageWidth = imageWidth
        self._numImages = numImages

    @property
    def imageWidth(self):
        """Scan image width"""
        return self._imageWidth

    @property
    def numImages(self):
        """Number of images"""
        return self._numImages


class PositionerSetting(IdentifiedElement):
    """Positioner Setting object.

    Has a uuid and a settings dictionary of axisName:value
    """

    def __init__(self, id_=None, **axisSettings):
        super().__init__(id_=id_)

        if self.__class__.__name__ == "PositionerSetting":
            # This class is abstract
            raise RuntimeError(
                "Programming error -"
                " attempt to instantiate abstract class PositionerSetting"
            )
        if None in axisSettings.values():
            raise ValueError(
                "axisSettings contain value None: %s" % sorted(axisSettings.items())
            )
        self._axisSettings = axisSettings.copy()

    @property
    def axisSettings(self):
        """axisName:value settings dictionary. NB the returned value is a copy;
        modifying it does *not* modify the object internals."""
        return self._axisSettings.copy()


class DetectorSetting(PositionerSetting):
    """Detector position setting"""


class BcsDetectorSetting(DetectorSetting):
    """Detector position setting with additional (beamline-side) resolution and orgxy"""

    def __init__(self, resolution, id_=None, orgxy=(), **axisSettings):
        super(BcsDetectorSetting, self).__init__(id_=id_, **axisSettings)
        self._resolution = resolution
        self._orgxy = tuple(orgxy)

    @property
    def resolution(self):
        """Resolution (in A) matching detector distance"""
        return self._resolution

    @property
    def orgxy(self):
        """Tuple, empty or of two floats; beam centre on detector"""
        return self._orgxy


class BeamstopSetting(PositionerSetting):
    """Beamstop position setting"""


class GoniostatRotation(PositionerSetting):
    """Goniostat Rotation setting"""

    def __init__(self, id_=None, **axisSettings):
        PositionerSetting.__init__(self, id_=id_, **axisSettings)
        self._translation = None

    @property
    def translation(self):
        """GoniostatTranslation corresponding to self

        NB This link can be set only by GoniostatTranslation.__init__"""
        return self._translation

    def get_motor_settings(self):
        """Get dictionary of rotation and translation motor setting"""
        result = dict(self.axisSettings)
        translation = self.translation
        if translation is not None:
            result.update(translation.axisSettings)
        return result


class GoniostatSweepSetting(GoniostatRotation):
    """Goniostat Sweep setting"""

    def __init__(self, scanAxis, id_=None, **axisSettings):
        GoniostatRotation.__init__(self, id_=id_, **axisSettings)
        self._scanAxis = scanAxis

    @property
    def scanAxis(self):
        """Scanning axis"""
        return self._scanAxis


class GoniostatTranslation(PositionerSetting):
    """Goniostat Translation setting

    NB the reverse GoniostatRotation.translation link is set from here.
    For this reason the constructor parameters are different from the
    object attributes. rotation is taken to be newRotation, except that:

    if ( rotation is not None and
          (requestedRotationId is None or requestedRotationId == rotation.id_):
        self.requestedRotationId = rotation.id_
        self.newRotation = None"""

    def __init__(
        self, rotation=None, requestedRotationId=None, id_=None, **axisSettings
    ):
        PositionerSetting.__init__(self, id_=id_, **axisSettings)

        if rotation is None:
            if requestedRotationId is None:
                raise ValueError("rotation and requestedRotationId cannot both be None")
            else:
                self._newRotation = None
                self._requestedRotationId = requestedRotationId
        else:
            if requestedRotationId is None or requestedRotationId == rotation.id_:
                self._newRotation = None
                self._requestedRotationId = rotation.id_
            else:
                self._newRotation = rotation
                self._requestedRotationId = requestedRotationId

            # NBNB this deliberately interferes with the internals of
            # GoniostatRotation
            rotation._translation = self

    @property
    def newRotation(self):
        return self._newRotation

    @property
    def requestedRotationId(self):
        return self._requestedRotationId


# Complex data objects


class UserProvidedInfo(MessageData):
    """User-provided information"""

    def __init__(self, data_model):
        self._scatterers = ()
        crystal_classes = data_model.crystal_classes
        if crystal_classes:
            self._crystalClasses = tuple(crystal_classes)
        else:
            self._crystalClasses = ()

        space_group = data_model.space_group
        sginfo = crystal_symmetry.SPACEGROUP_MAP.get(data_model.space_group)
        self._spaceGroup = sginfo.number if sginfo else None
        self._spaceGroupString = space_group or None
        cell_parameters = data_model.cell_parameters
        if cell_parameters:
            self._cell = UnitCell(*cell_parameters)
        else:
            self._cell = None
        detector_setting = data_model.detector_setting
        if detector_setting:
            self._expectedResolution = detector_setting.resolution
        else:
            self._expectedResolution = data_model.aimed_resolution
        self._isAnisotropic = None

    @property
    def scatterers(self):
        return self._scatterers

    @property
    def crystalClasses(self):
        return self._crystalClasses

    @property
    def spaceGroup(self):
        return self._spaceGroup

    @property
    def spaceGroupString(self):
        return self._spaceGroupString

    @property
    def cell(self):
        return self._cell

    @property
    def expectedResolution(self):
        return self._expectedResolution

    @property
    def isAnisotropic(self):
        return self._isAnisotropic


class Sweep(IdentifiedElement):
    """Geometric strategy Sweep"""

    def __init__(
        self,
        goniostatSweepSetting,
        detectorSetting,
        beamSetting,
        start,
        width,
        beamstopSetting=None,
        sweepGroup=None,
        id_=None,
    ):
        super(Sweep, self).__init__(id_=id_)

        self._scans = set()

        self._goniostatSweepSetting = goniostatSweepSetting
        self._detectorSetting = detectorSetting
        self._beamSetting = beamSetting
        self._beamstopSetting = beamstopSetting
        self._start = start
        self._width = width
        self._sweepGroup = sweepGroup

    @property
    def goniostatSweepSetting(self):
        return self._goniostatSweepSetting

    @property
    def detectorSetting(self):
        return self._detectorSetting

    @property
    def beamSetting(self):
        return self._beamSetting

    @property
    def beamstopSetting(self):
        return self._beamstopSetting

    @property
    def start(self):
        return self._start

    @property
    def width(self):
        return self._width

    @property
    def sweepGroup(self):
        return self._sweepGroup

    @property
    def scans(self):
        """Scans that belong to sweeps.

        NB this is a two-way link.
        It is populated when the Scan objects are created"""
        return frozenset(self._scans)

    def _add_scan(self, scan):
        """Implementation method. *Only* to be called from Scan.__init__"""
        self._scans.add(scan)

    def get_initial_settings(self):
        """Get dictionary of rotation and translation motor settings for start of sweep"""
        result = self.goniostatSweepSetting.get_motor_settings()
        result[self.goniostatSweepSetting.scanAxis] = self.start
        return result


class Scan(IdentifiedElement):
    """Collection strategy Scan"""

    def __init__(
        self, width, exposure, imageStartNum, start, sweep, filenameParams, id_=None
    ):
        super().__init__(id_=id_)

        self._filenameParams = dict(filenameParams)

        self._width = width
        self._exposure = exposure
        self._imageStartNum = imageStartNum
        self._start = start

        sweep._add_scan(self)
        self._sweep = sweep

    @property
    def width(self):
        return self._width

    @property
    def exposure(self):
        return self._exposure

    @property
    def imageStartNum(self):
        return self._imageStartNum

    @property
    def start(self):
        return self._start

    @property
    def sweep(self):
        return self._sweep

    @property
    def filenameParams(self):
        return dict(self._filenameParams)


class GeometricStrategy(IdentifiedElement, Payload):
    """Geometric strategy"""

    INTENT = "COMMAND"

    def __init__(
        self,
        isUserModifiable,
        defaultDetectorSetting,
        defaultBeamSetting,
        allowedWidths=(),
        sweepOffset=None,
        sweepRepeat=None,
        defaultWidthIdx=None,
        sweeps=(),
        id_=None,
    ):
        super().__init__(id_=id_)

        # self._isInterleaved = isInterleaved
        self._isUserModifiable = isUserModifiable
        self._defaultDetectorSetting = defaultDetectorSetting
        self._defaultBeamSetting = defaultBeamSetting
        self._defaultWidthIdx = defaultWidthIdx
        self._sweeps = frozenset(sweeps)

        if len(set(allowedWidths)) != len(allowedWidths):
            raise ValueError(
                "allowedWidths contains duplicate value: %s" % allowedWidths
            )
        else:
            self._allowedWidths = tuple(allowedWidths)

        self._sweepOffset = sweepOffset
        self._sweepRepeat = sweepRepeat

    @property
    def sweepRepeat(self):
        return self._sweepRepeat

    @property
    def sweepOffset(self):
        return self._sweepOffset

    @property
    def isUserModifiable(self):
        return self._isUserModifiable

    @property
    def defaultWidthIdx(self):
        return self._defaultWidthIdx

    @property
    def defaultDetectorSetting(self):
        return self._defaultDetectorSetting

    @property
    def defaultBeamSetting(self):
        return self._defaultBeamSetting

    @property
    def allowedWidths(self):
        return self._allowedWidths

    @property
    def sweeps(self):
        return self._sweeps

    def get_ordered_sweeps(self):
        """Get sweeps in acquisition order.

        Acquisition order is determined by the sweepGroup -
        to get results deterministic we use a secondary sort on
        angles, in alphabetical name order as a backup,
        (in pracite, 'kappa', 'kappa_phi', 'phi')
        which should match what the workflow does internally.
        Anyway, it is the sweep"""
        ll0 = []
        for sweep in self._sweeps:
            dd0 = sweep.get_initial_settings()
            ll0.append((sweep.sweepGroup, tuple(dd0[x] for x in sorted(dd0)), sweep))
        return list(tt0[2] for tt0 in sorted(ll0))


class CollectionProposal(IdentifiedElement, Payload):
    """Collection proposal"""

    INTENT = "COMMAND"

    def __init__(self, relativeImageDir, strategy, scans, id_=None):
        super().__init__(id_=id_)

        self._relativeImageDir = relativeImageDir
        self._strategy = strategy
        self._scans = tuple(scans)

    @property
    def relativeImageDir(self):
        return self._relativeImageDir

    @property
    def strategy(self):
        return self._strategy

    @property
    def scans(self):
        return self._scans


class PriorInformation(Payload):
    """Prior information to workflow calculation"""

    INTENT = "DOCUMENT"

    def __init__(self, data_model, image_root):
        super().__init__()

        # Look for existing uuid
        sample_model = data_model.get_sample_node()
        for text in sample_model.lims_code, sample_model.code, sample_model.name:
            if text:
                try:
                    sampleId = uuid.UUID(text)
                except:
                    # The error expected if this goes wrong is ValueError.
                    # But whatever the error we want to continue
                    pass
                else:
                    # Text was a valid uuid string. Use the uuid.
                    break
        else:
            sampleId = uuid.uuid1()
        if isinstance(sampleId, uuid.UUID):
            self._sampleId = sampleId
        else:
            raise TypeError("sampleId input must be of type uuid.UUID")

        self._sampleName = data_model.path_template.base_prefix
        # Draft in API, not currently coded in Java:
        # self._referenceFile = referenceFile
        self._rootDirectory = image_root
        self._userProvidedInfo = UserProvidedInfo(data_model)

    @property
    def sampleId(self):
        return self._sampleId

    @property
    def sampleName(self):
        return self._sampleName

    # @property
    # def referenceFile(self):
    #     return self._referenceFile

    @property
    def rootDirectory(self):
        return self._rootDirectory

    @property
    def userProvidedInfo(self):
        return self._userProvidedInfo


class RequestCentring(Payload):
    """Request for centering"""

    INTENT = "COMMAND"

    def __init__(self, currentSettingNo, totalRotations, goniostatRotation):
        super(RequestCentring, self).__init__()
        self._currentSettingNo = currentSettingNo
        self._totalRotations = totalRotations
        self._goniostatRotation = goniostatRotation

    @property
    def currentSettingNo(self):
        return self._currentSettingNo

    @property
    def totalRotations(self):
        return self._totalRotations

    @property
    def goniostatRotation(self):
        return self._goniostatRotation


class CentringDone(Payload):
    """Centering-done message"""

    INTENT = "DOCUMENT"

    def __init__(self, status, timestamp, goniostatTranslation):
        super().__init__()
        self._status = status
        self._timestamp = timestamp
        self._goniostatTranslation = goniostatTranslation

    @property
    def status(self):
        return self._status

    @property
    def timestamp(self):
        """Time in seconds since the epoch (Jan 1, 1970),
        as returned by time.time()"""
        return self._timestamp

    @property
    def goniostatTranslation(self):
        return self._goniostatTranslation


class SampleCentred(Payload):
    INTENT = "DOCUMENT"

    def __init__(self, data_model):
        super().__init__()
        self._imageWidth = data_model.image_width
        self._transmission = 0.01 * data_model.transmission
        self._exposure = data_model.exposure_time
        self._wedgeWidth = round(data_model.wedge_width / data_model.image_width)
        self._interleaveOrder = data_model.interleave_order
        self._beamstopSetting = data_model.beamstop_setting
        self._goniostatTranslations = frozenset(data_model.goniostat_translations)
        self._repetition_count = data_model.repetition_count

        self._detectorSetting = None
        if data_model.characterisation_done:
            self._wavelengths = tuple(data_model.wavelengths)
        else:
            # This trick assumes that characterisation and diffractcal
            # use one, the first, wavelength and default interleave order
            # Which is true. Not the ideal place to put this code,
            # but it works.
            self._wavelengths = tuple((data_model.wavelengths[0],))
            # if data_model.wftype != "diffractcal":
            #     self._detectorSetting = data_model.detector_setting
        if data_model.wftype != "diffractcal":
            self._detectorSetting = data_model.detector_setting

    @property
    def imageWidth(self):
        return self._imageWidth

    @property
    def transmission(self):
        return self._transmission

    @property
    def exposure(self):
        return self._exposure

    @property
    def interleaveOrder(self):
        return self._interleaveOrder

    @property
    def wedgeWidth(self):
        # Wedge width NB in *number of images*
        return self._wedgeWidth

    @property
    def beamstopSetting(self):
        return self._beamstopSetting

    @property
    def detectorSetting(self):
        return self._detectorSetting

    @property
    def goniostatTranslations(self):
        return self._goniostatTranslations

    @property
    def wavelengths(self):
        return self._wavelengths

    @property
    def repetition_count(self):
        return self._repetition_count



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
Detector API
Abstract methods:
    prepare_acquisition
    start_acquisition
    has_shutterless
Overloaded methods:
    force_emit_signals
Implemented methods:
    stop_acquisition
    get_pixel_size, get_width, get_heigth, get_metadata
    get_radius, get_outer_radius
    get_beam_position
    get_roi_mode, set_roi_mode, get_roi_mode_name, get_roi_modes
    get_exposure_time_limits
    get_threshold_energy, set_threshold_energy
    get_binning_mode, set_binning_mode
Implemented propertries:
    distance
Emitted signals:
    detectorRoiModeChanged
    temperatureChanged
    humidityChanged
    expTimeLimitsChanged
    frameRateChanged
    stateChanged
    specificStateChanged
Hardware objects used: energy
"""

import abc
import ast
import math

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.model.queue_model_objects import PathTemplate

__copyright__ = """ Copyright © 2019 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class AbstractDetector(HardwareObject):
    """Common base class for detectors"""

    __metaclass__ = abc.ABCMeta

    def __init__(self, name):
        super().__init__(name)

        self._temperature = None
        self._humidity = None
        self._actual_frame_rate = None
        self._pixel_size = (None, None)
        self._binning_mode = None
        self._roi_mode = 0
        self._images_per_file = 0
        self._roi_modes_list = []
        self._exposure_time_limits = (None, None)

        self._threshold_energy = None
        self._distance_motor_hwobj = None

        self._width = None  # [pixel]
        self._height = None  # [pixel]
        self._metadata = {}

    def init(self):
        """Initialise some common parameters"""
        super().init()

        self._metadata = self.get_property("beam", {})

        self._images_per_file = self.get_property("images_per_file", 100)

        self._distance_motor_hwobj = self.get_object_by_role("detector_distance")

        self._roi_modes_list = ast.literal_eval(self.get_property("roiModes", "()"))

        self._pixel_size = (self.get_property("px"), self.get_property("py"))
        self._width = self.get_property("width")
        self._height = self.get_property("height")

        min_exp_time = self.get_property("minimum_exposure_time", None)
        max_exp_time = self.get_property("maximum_exposure_time", None)

        if min_exp_time is None:
            self.log.warning("Minimum exposure time not set for detector, using None")
        else:
            min_exp_time = float(min_exp_time)

        if max_exp_time is None:
            self.log.warning("Maximum exposure time not set for detector, using None")
        else:
            max_exp_time = float(max_exp_time)

        self._exposure_time_limits = (
            min_exp_time,
            max_exp_time,
        )

    def force_emit_signals(self):
        """Emit all hardware object signals."""
        self.emit("detectorRoiModeChanged", (self._roi_mode,))
        self.emit("temperatureChanged", (self._temperature, True))
        self.emit("humidityChanged", (self._humidity, True))
        self.emit("expTimeLimitsChanged", (self._exposure_time_limits,))
        self.emit("frameRateChanged", (self._actual_frame_rate,))
        self.emit("stateChanged", (self._state,))
        self.emit("specificStateChanged", (self._specific_state,))

    @property
    def images_per_file(self):
        return self._images_per_file

    @property
    def distance(self):
        """Property for contained detector_distance hardware object
        Returns:
            (AbstratctMotor): Hardware object.
        """
        return self._distance_motor_hwobj

    @abc.abstractmethod
    def has_shutterless(self):
        """Check if detector is capable of shutterless acquisition.
        Returns:
            (bool): True if detector is capable, False otherwise
        """
        return

    @abc.abstractmethod
    def prepare_acquisition(self, *args, **kwargs):
        """
        Prepares detector for acquisition
        """

    def last_image_saved(self):
        """Get the path to the last saved image
        Returns:
            (str): full path.
        """
        return None

    @abc.abstractmethod
    def start_acquisition(self):
        """Start the acquisition."""

    def stop_acquisition(self):
        """Stop the acquisition."""

    def get_roi_mode(self):
        """Get the current ROI mode.
        Returns:
            (str): current ROI mode
        """
        return self._roi_mode

    def set_roi_mode(self, roi_mode):
        """Set the current ROI mode.
        Args:
            roi_mode (int): ROI mode to set.
        """
        self._roi_mode = roi_mode
        self.emit("detectorRoiModeChanged", (self._roi_mode,))

    def get_roi_mode_name(self):
        """Get the current ROI mode name.
        Returns:
            (str): name
        """
        return self._roi_modes_list[self._roi_mode]

    def get_roi_modes(self):
        """Get the available ROI modes
        Returns:
            (tuple): Tuple of strings.
        """
        return tuple(self._roi_modes_list)

    def get_exposure_time_limits(self):
        """Get the exposure time lower and upper limits.
        Returns:
            (tuple): Two floats (lower and upper limit) [s]
        """
        return self._exposure_time_limits

    def get_pixel_size(self):
        """Get the pixel size.
        Returns:
            (tuple): Two floats (x, y) [mm].
        """
        return self._pixel_size

    def get_binning_mode(self):
        """Get the current binning mode.
        Returns:
            (int): Binning mode
        """
        return self._binning_mode

    def set_binning_mode(self, value):
        """Set the current binning mode.
        Args:
            value (int): binning mode.
        """
        self._binning_mode = value

    def get_beam_position(self, distance=None, wavelength=None):  # noqa: ARG002
        """Calculate the beam position for a given distance.
        Args:
            distance (float): detector distance [mm]
            wavelength (float): X-ray wavelength [Å]
        Returns:
            tuple(float, float): Beam position x,y coordinates [pixel].
        """
        # Following is just an example of calculating beam center using
        # simple linear regression,
        # One needs to overload the method in case more complex calculation
        # is required, e.g. using other detector positioning motors and
        # wavelength

        try:
            distance = (
                distance
                if distance is not None
                else self._distance_motor_hwobj.get_value()
            )

            metadata = self.get_metadata()

            beam_position = (
                float(distance * metadata["ax"] + metadata["bx"]),
                float(distance * metadata["ay"] + metadata["by"]),
            )
        except (AttributeError, KeyError):
            beam_position = (None, None)

        return beam_position

    def get_radius(self, distance=None):
        """Get distance from the beam position to the nearest detector edge.
        Args:
            distance (float): Distance [mm]
        Returns:
            (float): Detector radius [mm]
        """
        try:
            distance = (
                distance
                if distance is not None
                else self._distance_motor_hwobj.get_value()
            )
        except AttributeError as err:
            raise RuntimeError("Cannot calculate radius, unknown distance") from err

        beam_x, beam_y = self.get_beam_position(distance)
        pixel_x, pixel_y = self.get_pixel_size()
        rrx = min(self.get_width() - beam_x, beam_x) * pixel_x
        rry = min(self.get_height() - beam_y, beam_y) * pixel_y
        radius = min(rrx, rry)

        return radius

    def get_outer_radius(self, distance=None):
        """Get distance from beam_position to the furthest point on the detector.
        Args:
            distance (float): Distance [mm]
        Returns:
            (float): Detector outer adius [mm]
        """
        try:
            distance = (
                distance
                if distance is not None
                else self._distance_motor_hwobj.get_value()
            )
        except AttributeError as err:
            raise RuntimeError(
                "Cannot calculate outer radius, distance unknown"
            ) from err

        beam_x, beam_y = self.get_beam_position(distance)
        pixel_x, pixel_y = self.get_pixel_size()
        max_delta_x = max(beam_x, self._width - beam_x) * pixel_x
        max_delta_y = max(beam_y, self._height - beam_y) * pixel_y

        return math.sqrt(max_delta_x * max_delta_x + max_delta_y * max_delta_y)

    def get_metadata(self):
        """Returns relevant metadata.
        Returns:
            (dict): metadata
        """
        self._metadata["width"] = self.get_width()
        self._metadata["height"] = self.get_height()

        return self._metadata

    def get_width(self):
        """Returns detector width.
        Returns:
            (int): detector width [px]
        """
        return self._width

    def get_height(self):
        """Returns detector height.
        Returns:
            (int): detector height [px]
        """
        return self._height

    def set_threshold_energy(self, threshold_energy):
        """
        Args:
            threshold_energy (float): Detector threshold energy [eV]
        """
        self._threshold_energy = threshold_energy

    def get_threshold_energy(self):
        """Returns detector threshold_energy
        Returns:
            (float): Detector threshold energy [eV]
        """
        return self._threshold_energy

    def get_image_file_name(self, path_template, suffix=None):
        template = "%s_%s_%%0" + str(path_template.precision) + "d.%s"
        suffix = suffix or path_template.suffix
        file_name = template % (
            path_template.get_prefix(),
            path_template.run_number,
            suffix,
        )
        if path_template.compression:
            file_name = "%s.gz" % file_name

        return file_name

    def get_first_and_last_file(self, pt: PathTemplate):
        """
        Get complete path to first and last image

        Args:
          pt (PathTempalte): Path template parameter

        Returns:
        (Tuple): Tuple containing first and last image path (first, last)
        """
        start_num = pt.start_num
        end_num = pt.start_num + pt.num_files - 1

        return (pt.get_image_path() % start_num, pt.get_image_path() % end_num)



# -*- coding: utf-8 -*-

from pydantic import (
    BaseModel,
    Field,
)


class ValidationError(Exception):
    pass


class BaseModel(BaseModel):
    def __init__(self, *args, **kwargs):
        try:
            super(BaseModel, self).__init__(*args, **kwargs)
        except Exception as ex:
            raise ValidationError(ex)


class MockDataModel(BaseModel):
    """
    Example DataModel used by the MockProcedure (for testing)
    """

    transmission: float = Field(0)
    energy: float = Field(0)
    resolution: float = Field(0)
    exposure_time: float = Field(0)
    number_of_images: float = Field(0)



import logging
import time

import gevent
from PX1Environment import EnvironmentPhase

from mxcubecore.Command.Tango import DeviceProxy
from mxcubecore.HardwareObjects.Cats90 import (
    BASKET_UNIPUCK,
    Cats90,
    SampleChangerState,
)


class PX1Cryotong(Cats90):
    __TYPE__ = "CATS"

    default_no_lids = 1
    baskets_per_lid = 3

    default_basket_type = BASKET_UNIPUCK

    def __init__(self, *args, **kwargs):
        super(PX1Cryotong, self).__init__(*args, **kwargs)

        self._safeNeeded = None
        self._homeOpened = None
        self.dry_and_soak_needed = False
        self.count_down = None

        self.soft_auth = None
        self.incoherent_state = None

    def init(self):
        super(PX1Cryotong, self).init()

        self.cats_device = DeviceProxy(self.get_property("cats_device"))

        self.environment = self.get_object_by_role("environment")

        if self.environment is None:
            logging.error(
                "PX1Cats. environment object not available. Sample changer cannot operate. Info.mode only"
            )
            self.infomode = True
        else:
            self.infomode = False

        for channel_name in (
            "_chnSoftAuth",
            "_chnHomeOpened",
            "_chnDryAndSoakNeeded",
            "_chnIncoherentGonioSampleState",
            "_chnSampleIsDetected",
            "_chnCountDown",
        ):
            setattr(self, channel_name, self.get_channel_object(channel_name))

        self._chnSoftAuth.connect_signal("update", self._software_authorization)
        self._chnHomeOpened.connect_signal("update", self._update_home_opened)
        self._chnIncoherentGonioSampleState.connect_signal(
            "update", self._update_ack_sample_memory
        )
        self._chnDryAndSoakNeeded.connect_signal("update", self._dry_and_soak_needed)
        self._chnSampleIsDetected.connect_signal(
            "update", self._update_sample_is_detected
        )
        self._chnCountDown.connect_signal("update", self._update_count_down)

        self._cmdDrySoak = self.add_command(
            {"type": "tango", "name": "_cmdDrySoak", "tangoname": self.tangoname},
            "DryAndSoak",
        )

    # ## CRYOTONG SPECIFIC METHODS ###

    def _software_authorization(self, value):
        if value != self.soft_auth:
            self.soft_auth = value
            self.emit("softwareAuthorizationChanged", (value,))

    def _update_home_opened(self, value=None):
        if self._homeOpened != value:
            self._homeOpened = value
            self.emit("homeOpened", (value,))

    def _update_sample_is_detected(self, value):
        self.emit("sampleIsDetected", (value,))

    def _update_ack_sample_memory(self, value=None):
        if value is None:
            value = self._chnIncoherentGonioSampleState.get_value()

        if value != self.incoherent_state:
            # automatically acknowledge the error. send a warning to the GUI
            if self.incoherent_state is not None:
                logging.getLogger("user_level_log").warning(
                    "CATS: Requested Sample could not be loaded."
                )
                self.emit("loadError", value)
                try:
                    self._cmdAckSampleMemory()
                except Exception:
                    """do nothing if cmd not to acknowledge not in xml"""
                    self.log.exception("")
            self.incoherent_state = value

    def _dry_and_soak_needed(self, value=None):
        self.dry_and_soak_needed = value

    def do_dry_and_soak(self):
        homeOpened = self._chnHomeOpened.get_value()

        if not homeOpened:
            self._do_dry_soak()
        else:
            logging.getLogger("user_level_log").warning(
                "CATS: You must Dry_and_Soak the gripper."
            )

    def _update_count_down(self, value=None):
        if value is None:
            value = self._chnCountDown.get_value()

        if value != self.count_down:
            self.log.info("PX1Cats. CountDown changed. Now is: %s" % value)
            self.count_down = value
            self.emit("countdownSignal", value)

    def _do_dry_soak(self):
        """
        Launch the "DrySoak" command on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        if self.infomode:
            logging.warning("PX1Cats. It is in info mode only. DrySoak command ignored")
            return

        self._cmdDrySoak()

    def _do_safe(self):
        """
        Launch the "safe" trajectory on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        if self.infomode:
            logging.warning(
                "PX1Cryotong. It is in info mode only. Command 'safe' ignored"
            )
            return

        ret = self.env_send_transfer()

        if not ret:
            logging.getLogger("user_level_log").error(
                "PX1 Environment cannot set transfer phase"
            )
            raise Exception(
                "Cryotong cannot get to transfer phase. Aborting sample changer operation"
            )

        self._execute_server_task(
            self._cmdSafe,
            "Safe",
            states=[SampleChangerState.Ready, SampleChangerState.Alarm],
        )

    # ## (END) CRYOTONG SPECIFIC METHODS ###

    # ## OVERLOADED CATS90 methods ####
    def cats_pathrunning_changed(self, value):
        Cats90.cats_pathrunning_changed(self, value)
        if self.cats_running is False and self.dry_and_soak_needed:
            self.do_dry_and_soak()

    def _do_load(self, sample=None, wash=None):
        ret = self.check_power_on()
        if ret is False:
            logging.getLogger("user_level_log").error("CRYOTONG Cannot be powered")
            raise Exception(
                "CRYOTONG Cannot be powered. Aborting sample changer operation"
            )

        ret = self.check_drysoak()
        if ret is False:
            logging.getLogger("user_level_log").error(
                "CRYOTONG Home Open / DryAndSoak not valid for loading"
            )
            raise Exception("CRYOTONG Home Open / DryAndSoak not valid for loading")

        ret = self.env_send_transfer()
        if ret is False:
            logging.getLogger("user_level_log").error(
                "PX1 Environment cannot set transfer phase"
            )
            raise Exception(
                "Cryotong cannot get to transfer phase. Aborting sample changer operation"
            )

        self._do_loadOperation(sample)

        # Check the value of the CATSCRYOTONG attribute dryAndSoakNeeded to warn
        # user if it is True
        dryAndSoak = self._chnDryAndSoakNeeded.get_value()
        if dryAndSoak:
            logging.getLogger("user_level_log").warning(
                "CATS: It is recommended to Dry_and_Soak the gripper."
            )

        incoherentSample = self._chnIncoherentGonioSampleState.get_value()
        if incoherentSample:
            logging.getLogger("user_level_log").info(
                "CATS: Load/Unload Error. Please try again."
            )
            self.emit("loadError", incoherentSample)

    def _do_unload(self, sample=None, wash=None):
        ret = self.check_power_on()
        if ret is False:
            logging.getLogger("user_level_log").error("CRYOTONG Cannot be powered")
            raise Exception(
                "CRYOTONG Cannot be powered. Aborting sample changer operation"
            )

        ret = self.env_send_transfer()

        if ret is False:
            logging.getLogger("user_level_log").error(
                "PX1 Environment cannot set transfer phase"
            )
            raise Exception(
                "Cryotong cannot get to transfer phase. Aborting sample changer operation"
            )

        self._do_unloadOperation(sample)

    def check_power_on(self):
        if self._chnPowered.get_value():
            return True

        self._cmdPowerOn()

        timeout = 3
        t0 = time.time()

        while not self._chnPowered.get_value():
            gevent.sleep(0.3)
            if time.time() - t0 > timeout:
                self.log.warning("CRYOTONG: timeout waiting for power on")
                break

        if self._chnPowered.get_value():
            return False

        return True

    def check_drysoak(self):
        if self._chnHomeOpened.get_value() is False:
            return True

        self._cmdDrySoak()

        time.sleep(3)
        t0 = time.time()
        wait_n = 0
        while self._is_device_busy():
            if wait_n % 10 == 3:
                self.log.warning("CRYOTONG: waiting for dry and soak to complete")
            gevent.sleep(0.3)
            wait_n += 1

        if self._is_device_ready() and self._chnHomeOpened.get_value() is False:
            return True
        else:
            return False

    def env_send_transfer(self):
        if self.environment.readyForTransfer():
            return True

        logging.getLogger("user_level_log").warning(
            "CRYOTONG: Not ready for transfer. sending it"
        )
        self.environment.setPhase(EnvironmentPhase.TRANSFER)

        timeout = 10
        t0 = time.time()
        while not self.environment.readyForTransfer():
            gevent.sleep(0.3)
            if time.time() - t0 > timeout:
                self.log.warning("CRYOTONG: timeout waiting for transfer phase")
                break
            self.log.warning("CRYOTONG: waiting for transfer phase to be set")

        if not self.environment.readyForTransfer():
            return False

        self.log.warning("CRYOTONG: ready for transfer now")
        return True

    # ## (END) OVERLOADED CATS90 methods ####


def test_hwo(hwo):
    import gevent

    basket_list = hwo.get_basket_list()
    sample_list = hwo.get_sample_list()
    print("Baskets/Samples in CATS: %s/%s" % (len(basket_list), len(sample_list)))
    gevent.sleep(2)
    sample_list = hwo.get_sample_list()
    print("No of samples is ", len(sample_list))

    for s in sample_list:
        if s.is_loaded():
            print("Sample %s loaded" % s.get_address())
            break

    if hwo.has_loaded_sample():
        print(
            "Currently loaded (%s): %s"
            % (hwo.has_loaded_sample(), hwo.get_loaded_sample().get_address())
        )

    print("\nCATS model is: ", hwo.cats_model)
    print("CATS state is: ", hwo.state)
    print("Sample on Magnet : ", hwo.cats_sample_on_diffr())
    print("All lids closed: ", hwo._chnAllLidsClosed.get_value())

    print("Sample Changer State is: ", hwo.get_status())
    for basketno in range(hwo.number_of_baskets):
        no = basketno + 1
        print("Tool for basket %d is: %d" % (no, hwo.tool_for_basket(no)))



# from MD2Motor import MD2Motor
import math

from mxcubecore.BaseHardwareObjects import HardwareObject


class MicrodiffApertureMockup(HardwareObject):
    def init(self):
        self.actuator_name = "CurrentApertureDiameter"
        self.motor_pos_attr_suffix = "Index"
        self._last_position_name = None

        self.predefined_position_attr = 1

        self.predefinedPositions = {"5": 1, "20": 2, "50": 3, "100": 4, "150": 5}
        self.sortPredefinedPositionsList()

    def sortPredefinedPositionsList(self):
        self.predefinedPositionsNamesList = self.predefinedPositions.keys()
        self.predefinedPositionsNamesList.sort(
            lambda x, y: int(
                round(self.predefinedPositions[x] - self.predefinedPositions[y])
            )
        )

    def connect_notify(self, signal):
        if signal == "predefinedPositionChanged":
            positionName = self.get_current_position_name()

            try:
                pos = self.predefinedPositions[positionName]
            except KeyError:
                self.emit(signal, ("", None))
            else:
                self.emit(signal, (positionName, pos))
        else:
            return True

    def get_state(self):
        return 2

    def get_limits(self):
        return (1, 5)

    def get_predefined_positions_list(self):
        return self.predefinedPositionsNamesList

    def motor_positions_changed(self, absolutePosition, private={}):
        # MD2Motor.motor_positions_changed.im_func(self, absolutePosition, private)
        positionName = self.get_current_position_name(absolutePosition)
        if self._last_position_name != positionName:
            self._last_position_name = positionName
            self.emit(
                "predefinedPositionChanged",
                (positionName, positionName and absolutePosition or None),
            )

    def get_current_position_name(self, pos=None):
        pos = self.predefined_position_attr

        for positionName in self.predefinedPositions:
            if math.fabs(self.predefinedPositions[positionName] - pos) <= 1e-3:
                return positionName
        return ""

    def moveToPosition(self, positionName):
        try:
            self.predefined_position_attr = self.predefinedPositions[positionName]
            return True
        except Exception:
            return False



#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.


from mxcubecore import HardwareRepository as HWR
from mxcubecore.queue_entry.base_queue_entry import BaseQueueEntry

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3+"
__category__ = "General"


class OpticalCentringQueueEntry(BaseQueueEntry):
    """
    Entry for automatic sample centring with lucid
    """

    def __init__(self, view=None, data_model=None):
        BaseQueueEntry.__init__(self, view, data_model)

    def execute(self):
        BaseQueueEntry.execute(self)
        HWR.beamline.diffractometer.automatic_centring_try_count = (
            self.get_data_model().try_count
        )

        HWR.beamline.diffractometer.start_centring_method(
            HWR.beamline.diffractometer.CENTRING_METHOD_AUTO, wait=True
        )

    def pre_execute(self):
        BaseQueueEntry.pre_execute(self)

    def post_execute(self):
        self.get_view().set_checkable(False)
        BaseQueueEntry.post_execute(self)

    def get_type_str(self):
        return "Optical automatic centering"



import abc
import logging
import os
import time

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.TaskUtils import error_cleanup


class AbstractEnergyScan(HardwareObject):
    """Energy Scan abstract class"""

    __metaclass__ = abc.ABCMeta

    def __init__(self, name):
        super().__init__(name)
        self.data_collect_task = None
        self._egyscan_task = None
        self.scanning = False
        self.cpos = None
        self.energy_scan_parameters = {}

    def init(self):
        """Initialisation"""
        self.lims = HWR.beamline.lims
        if not self.lims:
            logging.getLogger().warning("EnergyScan: no lims set")

    def get_elements(self):
        """Get the configured in the file elements to be used
        Returns:
            (dict): Dictionary {"symbol": str, "energy": str}
        """
        try:
            return self.get_property("elements")["element"]
        except TypeError:
            return []

    def open_fast_shutter(self):
        """
        Open the fast shutter.
        """

    def close_fast_shutter(self):
        """
        Close the fast shutter.
        """

    def energy_scan_hook(self, energy_scan_parameters):
        """
        Execute actions, required before running the raw scan(like changing
        undulator gaps, move to a given energy... These are in general
        beamline specific actions.
        """

    def execute_energy_scan(self, energy_scan_parameters):
        """
        Execute the raw scan sequence. Here is where you pass whatever
        parameters you need to run the raw scan (e.g start/end energy,
        counting time, energy step...).
        """

    def get_static_parameters(self, config_file, element, edge):
        """
        Get any parameters, which are known before hand. Some of them are
        known from the theory, like the peak energy, others are equipment
        specific like the lower/upper ROI limits of the fluorescence detector.
        Usually these parameters are pre-defined in a file, but can also be
        calculated.
        The function should return a distionary with at least defined
        {'edgeEnergy': peak_energy} member, where 'edgeEnergy' is a
        compulsory key.
        It is convenient to put in the same dictionary the remote energy,
        the ROI min/max values.
        There are few more reserved key names:
        'eroi_min', 'eroi_max' - min and max ROI limits if you want to set one.
        'findattEnergy' - energy to move to if you want to choose the attenuation
        for the scan.
        """
        return {}

    def set_mca_roi(self, eroi_min, eroi_max):
        """
        Configure the fluorescent detector ROI. The input is min/max energy.
        """

    def escan_prepare(self):
        """
        Set the nesessary equipment in position for the scan.
        No need to know the scan paramets.
        """

    def choose_attenuation(self):
        """
        Procedure to set the minimal attenuation in order no preserve
        the sample. Should be done at the energy after the edge.
        """

    def escan_cleanup(self):
        """Execute actions at the end of the scan"""

    def escan_postscan(self):
        """
        set the nesessary equipment in position after the scan
        """

    def do_energy_scan(self):
        """Execute the scan"""
        with error_cleanup(self.escan_cleanup):
            self.escan_prepare()
            self.energy_scan_hook(self.energy_scan_parameters)
            HWR.beamline.safety_shutter.open(timeout=10)
            self.choose_attenuation()
            self.close_fast_shutter()
            self.log.debug("Doing the scan, please wait...")
            self.execute_energy_scan(self.energy_scan_parameters)
            self.escan_postscan()
            self.close_fast_shutter()
            self.energy_scan_parameters["flux"] = HWR.beamline.flux.get_value()
            HWR.beamline.safety_shutter.close(timeout=10)
            # send finish successfully signal
            self.emit("energyScanFinished", (self.energy_scan_parameters,))
            self.ready_event.set()

    def start_energy_scan(
        self,
        element,
        edge,
        directory,
        prefix,
        session_id=None,
        blsample_id=None,
        cpos=None,
    ):
        """Do the scan"""
        if self._egyscan_task and not self._egyscan_task.ready():
            raise RuntimeError("Scan already started.")

        self.emit("energyScanStarted", ())
        # Set the energy from the element and edge parameters
        static_pars_dict = {}
        static_pars_dict = self.get_static_parameters(
            self.get_property("config_file"), element, edge
        )
        self.cpos = cpos
        self.energy_scan_parameters = static_pars_dict
        self.energy_scan_parameters["element"] = element
        self.energy_scan_parameters["edge"] = edge
        self.energy_scan_parameters["directory"] = directory

        # Calculate the MCA ROI (if needed)
        try:
            self.set_mca_roi(static_pars_dict["eroi_min"], static_pars_dict["eroi_max"])
        except Exception:
            pass

        # create the directory if needed
        if not os.path.exists(directory):
            os.makedirs(directory)
        self.energy_scan_parameters["prefix"] = prefix
        if session_id is not None:
            self.energy_scan_parameters["sessionId"] = session_id
            self.energy_scan_parameters["blSampleId"] = blsample_id
            self.energy_scan_parameters["startTime"] = time.strftime(
                "%Y-%m-%d %H:%M:%S"
            )

        self._egyscan_task = gevent.spawn(self.do_energy_scan)

        """
        Use chooch to calculate edge and inflection point
        The brick expects the following parameters to be returned:
        pk, fppPeak, fpPeak, ip, fppInfl, fpInfl, rm,
        chooch_graph_x, chooch_graph_y1, chooch_graph_y2, title)
        """



#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging

import gevent
from helical_scan import helical_scan
from omega_scan import omega_scan

# from xray_centring import xray_centring
from raster_scan import raster_scan
from reference_images import reference_images
from slits import slits1

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects.abstract.AbstractCollect import AbstractCollect
from mxcubecore.TaskUtils import task

__credits__ = ["Synchrotron SOLEIL"]
__version__ = "2.3."
__category__ = "General"


class PX2Collect(AbstractCollect, HardwareObject):
    """Main data collection class. Inherited from AbstractCollect.
    Collection is done by setting collection parameters and
    executing collect command
    """

    experiment_types = [
        "omega_scan",
        "reference_images",
        "inverse_scan",
        "mad",
        "helical_scan",
        "xrf_spectrum",
        "energy_scan",
        "raster_scan",
        "nested_helical_acquisition",
        "tomography",
        "film",
        "optical_centering",
    ]

    # experiment_types = ['OSC',
    # 'Collect - Multiwedge',
    # 'Helical',
    # 'Mesh',
    # 'energy_scan',
    # 'xrf_spectrum',
    # 'neha'

    def __init__(self, name):
        """
        :param name: name of the object
        :type name: string
        """

        AbstractCollect.__init__(self, name)
        HardwareObject.__init__(self, name)

        self.current_dc_parameters = None
        self.osc_id = None
        self.owner = None
        self.aborted_by_user = None
        self.slits1 = slits1()

    def init(self):
        self.ready_event = gevent.event.Event()

        undulators = []
        try:
            for undulator in self["undulators"]:
                undulators.append(undulator)
        except Exception:
            self.log.exception("")

        beam_div_hor, beam_div_ver = HWR.beamline.beam.get_beam_divergence()

        self.set_beamline_configuration(
            synchrotron_name="SOLEIL",
            directory_prefix=self.get_property("directory_prefix"),
            default_exposure_time=HWR.beamline.detector.get_property(
                "default_exposure_time"
            ),
            minimum_exposure_time=HWR.beamline.detector.get_property(
                "minimum_exposure_time"
            ),
            detector_fileext=HWR.beamline.detector.get_property("fileSuffix"),
            detector_type=HWR.beamline.detector.get_property("type"),
            detector_manufacturer=HWR.beamline.detector.get_property("manufacturer"),
            detector_model=HWR.beamline.detector.get_property("model"),
            detector_px=HWR.beamline.detector.get_property("px"),
            detector_py=HWR.beamline.detector.get_property("py"),
            undulators=undulators,
            focusing_optic=self.get_property("focusing_optic"),
            monochromator_type=self.get_property("monochromator"),
            beam_divergence_vertical=beam_div_ver,
            beam_divergence_horizontal=beam_div_hor,
            polarisation=self.get_property("polarisation"),
            input_files_server=self.get_property("input_files_server"),
        )

        self.emit("collectConnected", (True,))
        self.emit("collectReady", (True,))

    def data_collection_hook(self):
        """Main collection hook"""

        if self.aborted_by_user:
            self.collection_failed("Aborted by user")
            self.aborted_by_user = False
            return

        parameters = self.current_dc_parameters

        log = logging.getLogger("user_level_info")
        log.info("data collection parameters received %s" % parameters)

        for parameter in parameters:
            log.info("%s: %s" % (str(parameter), str(parameters[parameter])))

        osc_seq = parameters["oscillation_sequence"][0]
        fileinfo = parameters["fileinfo"]
        sample_reference = parameters["sample_reference"]
        experiment_type = parameters["experiment_type"]
        energy = parameters["energy"]
        transmission = parameters["transmission"]
        resolution = parameters["resolution"]

        exposure_time = osc_seq["exposure_time"]
        in_queue = parameters["in_queue"] != False

        overlap = osc_seq["overlap"]
        angle_per_frame = osc_seq["range"]
        scan_start_angle = osc_seq["start"]
        number_of_images = osc_seq["number_of_images"]
        image_nr_start = osc_seq["start_image_number"]

        directory = fileinfo["directory"]
        prefix = fileinfo["prefix"]
        template = fileinfo["template"]
        run_number = fileinfo["run_number"]
        process_directory = fileinfo["process_directory"]

        # space_group = str(sample_reference['space_group'])
        # unit_cell = list(eval(sample_reference['cell']))

        self.emit("collectStarted", (self.owner, 1))
        self.emit("progressInit", ("Data collection", 100))
        self.emit("fsmConditionChanged", "data_collection_started", True)

        self.store_image_in_lims_by_frame_num(1)

        name_pattern = template[:-8]

        if experiment_type == "OSC":
            scan_range = angle_per_frame * number_of_images
            scan_exposure_time = exposure_time * number_of_images
            experiment = omega_scan(
                name_pattern,
                directory,
                scan_range=scan_range,
                scan_exposure_time=scan_exposure_time,
                scan_start_angle=scan_start_angle,
                angle_per_frame=angle_per_frame,
                image_nr_start=image_nr_start,
                photon_energy=energy,
                transmission=transmission,
                resolution=resolution,
                simulation=False,
            )
            experiment.execute()

        elif experiment_type == "Characterization":
            number_of_wedges = osc_seq["number_of_images"]
            wedge_size = osc_seq["wedge_size"]
            overlap = osc_seq["overlap"]
            scan_start_angles = []
            scan_exposure_time = exposure_time * wedge_size
            scan_range = angle_per_frame * wedge_size

            for k in range(number_of_wedges):
                scan_start_angles.append(
                    scan_start_angle + k * -overlap + k * scan_range
                )

            experiment = reference_images(
                name_pattern,
                directory,
                scan_range=scan_range,
                scan_exposure_time=scan_exposure_time,
                scan_start_angles=scan_start_angles,
                angle_per_frame=angle_per_frame,
                image_nr_start=image_nr_start,
                photon_energy=energy,
                transmission=transmission,
                resolution=resolution,
                simulation=False,
            )

            experiment.execute()

        elif experiment_type == "Helical" and osc_seq["mesh_range"] == ():
            scan_range = angle_per_frame * number_of_images
            scan_exposure_time = exposure_time * number_of_images
            log.info("helical_pos %s" % self.helical_pos)
            experiment = helical_scan(
                name_pattern,
                directory,
                scan_range=scan_range,
                scan_exposure_time=scan_exposure_time,
                scan_start_angle=scan_start_angle,
                angle_per_frame=angle_per_frame,
                image_nr_start=image_nr_start,
                position_start=self.translate_position(self.helical_pos["1"]),
                position_end=self.translate_position(self.helical_pos["2"]),
                photon_energy=energy,
                transmission=transmission,
                resolution=resolution,
                simulation=False,
            )
            experiment.execute()

        elif experiment_type == "Helical" and osc_seq["mesh_range"] != ():
            horizontal_range, vertical_range = osc_seq["mesh_range"]

            experiment = xray_centring(name_pattern, directory)

            experiment.execute(simulation=False)

        elif experiment_type == "Mesh":
            number_of_columns = osc_seq["number_of_lines"]
            number_of_rows = int(number_of_images / number_of_columns)
            horizontal_range, vertical_range = osc_seq["mesh_range"]
            angle_per_line = angle_per_frame * number_of_columns
            experiment = raster_scan(
                name_pattern,
                directory,
                vertical_range,
                horizontal_range,
                number_of_rows,
                number_of_columns,
                frame_time=exposure_time,
                scan_start_angle=scan_start_angle,
                scan_range=angle_per_line,
                image_nr_start=image_nr_start,
                photon_energy=energy,
                transmission=transmission,
                simulation=False,
            )
            experiment.execute()

        # for image in range(number_of_images):
        # if self.aborted_by_user:
        # self.ready_event.set()
        # return

        # Uncomment to test collection failed
        # if image == 5:
        # self.emit("collectOscillationFailed", (self.owner, False,
        # "Failed on 5", parameters.get("collection_id")))
        # self.ready_event.set()
        # return

        # gevent.sleep(exposure_time)
        # self.emit("collectImageTaken", image)
        # self.emit("progressStep", (int(float(image) / number_of_images * 100)))

        self.emit_collection_finished()

    def translate_position(self, position):
        translation = {
            "sampx": "CentringX",
            "sampy": "CentringY",
            "phix": "AlignmentX",
            "phiy": "AlignmentY",
            "phiz": "AlignmentZ",
        }
        translated_position = {}
        for key in position:
            if key in translation:
                translated_position[translation[key]] = position[key]
            else:
                translated_position[key] = position[key]
        return translated_position

    def trigger_auto_processing(self, process_event, params_dict, frame_number):
        """
        Descript. :
        """
        if HWR.beamline.offline_processing is not None:
            HWR.beamline.offline_processing.execute_autoprocessing(
                process_event,
                self.current_dc_parameters,
                frame_number,
                self.run_offline_processing,
            )

    @task
    def _take_crystal_snapshot(self, filename):
        HWR.beamline.sample_view.save_snapshot(filename)

    @task
    def _take_crystal_animation(self, animation_filename, duration_sec):
        """Rotates sample by 360 and composes a gif file
        Animation is saved as the fourth snapshot
        """
        HWR.beamline.sample_view.save_scene_animation(animation_filename, duration_sec)

    @task
    def move_motors(self, motor_position_dict):
        """
        Descript. :
        """
        return

    def emit_collection_finished(self):
        """Collection finished behaviour"""
        if self.current_dc_parameters["experiment_type"] != "Collect - Multiwedge":
            self.update_data_collection_in_lims()

            last_frame = self.current_dc_parameters["oscillation_sequence"][0][
                "number_of_images"
            ]
            if last_frame > 1:
                self.store_image_in_lims_by_frame_num(last_frame)
            if (
                self.current_dc_parameters["experiment_type"] in ("OSC", "Helical")
                and self.current_dc_parameters["oscillation_sequence"][0]["overlap"]
                == 0
                and last_frame > 19
            ):
                self.trigger_auto_processing("after", self.current_dc_parameters, 0)

        success_msg = "Data collection successful"
        self.current_dc_parameters["status"] = success_msg
        self.emit(
            "collectOscillationFinished",
            (
                self.owner,
                True,
                success_msg,
                self.current_dc_parameters.get("collection_id"),
                self.osc_id,
                self.current_dc_parameters,
            ),
        )
        self.emit("collectEnded", self.owner, success_msg)
        self.emit("collectReady", (True,))
        self.emit("progressStop", ())
        self.emit("fsmConditionChanged", "data_collection_successful", True)
        self.emit("fsmConditionChanged", "data_collection_started", False)
        self._collecting = None
        self.ready_event.set()

    def store_image_in_lims_by_frame_num(self, frame, motor_position_id=None):
        """
        Descript. :
        """
        image_id = None
        self.trigger_auto_processing("image", self.current_dc_parameters, frame)
        image_id = self.store_image_in_lims(frame)
        return image_id

    def stopCollect(self, owner="MXCuBE"):
        """
        Descript. :
        """
        self.aborted_by_user = True
        self.cmd_collect_abort()
        self.emit_collection_failed("Aborted by user")

    def set_helical_pos(self, helical_pos):
        self.helical_pos = helical_pos

    def get_slit_gaps(self):
        return self.get_slits_gap()

    def get_slits_gap(self):
        return self.slits1.get_horizontal_gap(), self.slits1.get_vertical_gap()



"""Command package.

The Command package is the place to put Command launchers
and Channel readers/writers modules.

The modules are organised by Control System Software used at different facilities
or specific to the hardware targeted.
Not all the facilities or even different beamlines from a facility have the same dependencies installed.
**One should modify the** ``pyproject.toml`` **to address their needs.**

CommandObject class and ChannelObject class derive from the mxcubecore.CommandContainer module.
They should both emit the following signals :
    - ``connected``, when connected to the control software
    - ``disconnected``, when disconnected from control software
They should both implement the ``is_connected()`` method.

    - Command launcher class in the module should emit the following signals :
        - ``commandBeginWaitReply``, when the command has been sent and we are waiting for the reply.

        - ``commandReplyArrived``, when the reply for the commandis arrived.

        - ``commandFailed``, when the command failed to execute.

        - ``commandAborted``, when the command has been aborted.

Every Command launcher should be a callable object ;
the arguments provided are those for the remote command to
be executed.

    - A Channel object should:
        - have a ``get_value`` method
        - have a ``set_value`` method (optional)
        - emit an ``update`` signal when the value of the channel changes
"""



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """Copyright The MXCuBE Collaboration"""
__license__ = "LGPLv3+"

import logging
import ssl
from urllib.error import URLError

from suds import WebFault
from suds.transport import TransportError

from mxcubecore import HardwareRepository as HWR
from mxcubecore.mxcubecore.HardwareObjects.ProposalTypeISPyBLims import (
    ProposalISPyBClient,
)

ssl._create_default_https_context = ssl._create_unverified_context


class P11ISPyBClient(ProposalISPyBClient):
    def init(self):
        ISPyBClient.init(self)

        self.simulated_proposal = self.get_property("proposal_simulated")

        if self.simulated_proposal == 1:
            self.simulated_prop_code = self.get_property("proposal_code_simulated")
            self.simulated_prop_number = self.get_property("proposal_number_simulated")
        else:
            self.simulated_prop_code = None
            self.simulated_prop_number = None
        self.log.debug("PROPOSAL SIMULATED is %s" % self.simulated_proposal)
        self.log.debug("PROPOSAL CODE is %s" % self.simulated_prop_code)
        self.log.debug("PROPOSAL NUMBER is %s" % self.simulated_prop_number)

    def update_data_collection(self, mx_collection, wait=False):
        mx_collection["beamline_name"] = "P11"
        ISPyBClient.update_data_collection(self, mx_collection, wait)

    def _store_data_collection(self, mx_collection, bl_config=None):
        self.prepare_collect_for_lims(mx_collection)
        return ISPyBClient._store_data_collection(self, mx_collection, bl_config)

    def store_image(self, image_dict):
        self.prepare_image_for_lims(image_dict)
        return ISPyBClient.store_image(self, image_dict)

    def store_robot_action(self, robot_action_dict):
        # TODO ISPyB is not ready for now. This prevents from error 500 from the server.
        pass

    def _store_data_collection_group(self, group_data):
        """ """

        # Workaround to make the data collection work even if ISPyB is not available due to the beamtime is not opened.
        try:
            group_id = self._collection.service.storeOrUpdateDataCollectionGroup(
                group_data
            )
        except:
            group_id = -9999

        return group_id

    def prepare_collect_for_lims(self, mx_collect_dict):
        # Attention! directory passed by reference. modified in place

        prop = "EDNA_files_dir"
        path = mx_collect_dict[prop]
        ispyb_path = HWR.beamline.session.path_to_ispyb(path)
        mx_collect_dict[prop] = ispyb_path

        prop = "process_directory"
        path = mx_collect_dict["fileinfo"][prop]
        ispyb_path = HWR.beamline.session.path_to_ispyb(path)
        mx_collect_dict["fileinfo"][prop] = ispyb_path

        for i in range(4):
            try:
                prop = "xtalSnapshotFullPath%d" % (i + 1)
                path = mx_collect_dict[prop]
                ispyb_path = HWR.beamline.session.path_to_ispyb(path)
                logging.debug("P11 ISPyBClient - %s is %s " % (prop, ispyb_path))
                mx_collect_dict[prop] = ispyb_path
            except RuntimeWarning("Can not get ISPyB path for %s" % prop):
                self.log.exception("")

    def prepare_image_for_lims(self, image_dict):
        for prop in ["jpegThumbnailFileFullPath", "jpegFileFullPath"]:
            try:
                path = image_dict[prop]
                ispyb_path = HWR.beamline.session.path_to_ispyb(path)
                image_dict[prop] = ispyb_path
            except RuntimeWarning("Can not prepare image path fir LIMS for %s" % prop):
                self.log.exception("")

    def get_proposal(self, proposal_code, proposal_number):
        self.log.debug(
            "ISPyB. Obtaining proposal for code=%s / prop_number=%s"
            % (proposal_code, proposal_number)
        )

        try:
            if self._shipping:
                # Attempt to fetch the proposal from ISPyB
                proposal = self._shipping.service.findProposal(
                    proposal_code, proposal_number
                )
            else:
                raise URLError("Shipping service unavailable")

            if proposal:
                proposal["code"] = proposal_code
                proposal["number"] = proposal_number
                return {"Proposal": proposal, "status": {"code": "ok"}}
        except (WebFault, URLError, TransportError) as e:
            # Log the error and fallback
            logging.getLogger("ispyb_client").exception(
                "Error fetching proposal. Returning fallback values."
            )
            return {
                "Proposal": {
                    "code": proposal_code,
                    "number": proposal_number,
                    "title": "Unknown Proposal",
                },
                "status": {"code": "error", "msg": "ISPyB is not connected."},
            }



#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.dispatcher import dispatcher
from mxcubecore.model import queue_model_objects
from mxcubecore.model.queue_model_enumerables import (
    COLLECTION_ORIGIN_STR,
    EXPERIMENT_TYPE,
)
from mxcubecore.queue_entry.base_queue_entry import (
    QUEUE_ENTRY_STATUS,
    BaseQueueEntry,
    QueueAbortedException,
    QueueExecutionException,
    center_before_collect,
)

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3+"
__category__ = "General"


class DataCollectionQueueEntry(BaseQueueEntry):
    """
    Defines the behaviour of a data collection.
    """

    def __init__(self, view=None, data_model=None, view_set_queue_entry=True):
        BaseQueueEntry.__init__(self, view, data_model, view_set_queue_entry)

        self.collect_task = None
        self.centring_task = None
        self.in_queue = False

    def __setstate__(self, d):
        self.__dict__.update(d)

    def __getstate__(self):
        d = dict(self.__dict__)
        d["collect_task"] = None
        d["centring_task"] = None
        d["shape_history"] = (
            HWR.beamline.sample_view.name() if HWR.beamline.sample_view else None
        )
        d["session"] = HWR.beamline.session.name if HWR.beamline.session else None
        d["lims_client_hwobj"] = HWR.beamline.lims.name if HWR.beamline.lims else None
        return d

    def execute(self):
        BaseQueueEntry.execute(self)
        data_collection = self.get_data_model()

        if data_collection:
            acq_params = data_collection.acquisitions[0].acquisition_parameters
            cpos = acq_params.centred_position

            empty_cpos = all(mpos is None for mpos in cpos.as_dict().values())

            if empty_cpos and data_collection.center_before_collect:
                _p, _s = center_before_collect(
                    self.get_view(),
                    HWR.beamline.diffractometer,
                    self.get_queue_controller(),
                    HWR.beamline.sample_view,
                )

                acq_params.centred_position = _p

            self.collect_dc(data_collection, self.get_view())

        if HWR.beamline.sample_view:
            HWR.beamline.sample_view.de_select_all()

    def pre_execute(self):
        BaseQueueEntry.pre_execute(self)

        qc = self.get_queue_controller()

        qc.connect(HWR.beamline.collect, "collectStarted", self.collect_started)
        qc.connect(
            HWR.beamline.collect, "collectNumberOfFrames", self.preparing_collect
        )
        qc.connect(
            HWR.beamline.collect, "collectOscillationStarted", self.collect_osc_started
        )
        qc.connect(
            HWR.beamline.collect, "collectOscillationFailed", self.collect_failed
        )
        qc.connect(
            HWR.beamline.collect, "collectOscillationFinished", self.collect_finished
        )
        qc.connect(HWR.beamline.collect, "collectImageTaken", self.image_taken)
        qc.connect(
            HWR.beamline.collect, "collectNumberOfFrames", self.collect_number_of_frames
        )

        if HWR.beamline.online_processing is not None:
            qc.connect(
                HWR.beamline.online_processing,
                "processingFinished",
                self.online_processing_finished,
            )
            qc.connect(
                HWR.beamline.online_processing,
                "processingFailed",
                self.online_processing_failed,
            )

        data_model = self.get_data_model()

        if data_model.get_parent():
            gid = data_model.get_parent().lims_group_id
            data_model.lims_group_id = gid

    def post_execute(self):
        BaseQueueEntry.post_execute(self)
        qc = self.get_queue_controller()

        qc.disconnect(HWR.beamline.collect, "collectStarted", self.collect_started)
        qc.disconnect(
            HWR.beamline.collect, "collectNumberOfFrames", self.preparing_collect
        )
        qc.disconnect(
            HWR.beamline.collect, "collectOscillationStarted", self.collect_osc_started
        )
        qc.disconnect(
            HWR.beamline.collect, "collectOscillationFailed", self.collect_failed
        )
        qc.disconnect(
            HWR.beamline.collect, "collectOscillationFinished", self.collect_finished
        )
        qc.disconnect(HWR.beamline.collect, "collectImageTaken", self.image_taken)
        qc.disconnect(
            HWR.beamline.collect, "collectNumberOfFrames", self.collect_number_of_frames
        )

        if HWR.beamline.online_processing is not None:
            qc.disconnect(
                HWR.beamline.online_processing,
                "processingFinished",
                self.online_processing_finished,
            )
            qc.disconnect(
                HWR.beamline.online_processing,
                "processingFailed",
                self.online_processing_failed,
            )

        self.get_view().set_checkable(False)

    def collect_dc(self, dc, list_item):
        log = logging.getLogger("user_level_log")

        if HWR.beamline.collect:
            acq_1 = dc.acquisitions[0]
            acq_1.acquisition_parameters.in_queue = self.in_queue
            cpos = acq_1.acquisition_parameters.centred_position
            sample = self.get_data_model().get_sample_node()
            HWR.beamline.collect.run_offline_processing = dc.run_offline_processing
            HWR.beamline.collect.aborted_by_user = None
            self.online_processing_task = None

            try:
                if dc.experiment_type is EXPERIMENT_TYPE.HELICAL:
                    acq_1, acq_2 = (dc.acquisitions[0], dc.acquisitions[1])
                    HWR.beamline.collect.set_helical(True)
                    HWR.beamline.collect.set_mesh(False)
                    HWR.beamline.collect.set_fast_characterisation(False)
                    start_cpos = acq_1.acquisition_parameters.centred_position
                    end_cpos = acq_2.acquisition_parameters.centred_position
                    helical_oscil_pos = {
                        "1": start_cpos.as_dict(),
                        "2": end_cpos.as_dict(),
                    }
                    HWR.beamline.collect.set_helical_pos(helical_oscil_pos)
                    # msg = "Helical data collection, moving to start position"
                    # log.info(msg)
                    # list_item.setText(1, "Moving sample")
                elif dc.experiment_type is EXPERIMENT_TYPE.MESH:
                    mesh_nb_lines = acq_1.acquisition_parameters.num_lines
                    mesh_total_nb_frames = acq_1.acquisition_parameters.num_images
                    mesh_range = acq_1.acquisition_parameters.mesh_range
                    mesh_center = acq_1.acquisition_parameters.centred_position
                    HWR.beamline.collect.set_mesh_scan_parameters(
                        mesh_nb_lines, mesh_total_nb_frames, mesh_center, mesh_range
                    )
                    HWR.beamline.collect.set_helical(False)
                    HWR.beamline.collect.set_fast_characterisation(False)
                    HWR.beamline.collect.set_mesh(True)
                    # inc_used_for_collection does nothing
                    HWR.beamline.sample_view.inc_used_for_collection(
                        self.get_data_model().shape
                    )
                elif dc.experiment_type is EXPERIMENT_TYPE.EDNA_REF:
                    HWR.beamline.collect.set_helical(False)
                    HWR.beamline.collect.set_mesh(False)
                    HWR.beamline.collect.set_fast_characterisation(True)
                else:
                    HWR.beamline.collect.set_helical(False)
                    HWR.beamline.collect.set_mesh(False)
                    HWR.beamline.collect.set_fast_characterisation(False)
                if (
                    dc.run_online_processing
                    and acq_1.acquisition_parameters.num_images > 4
                    and HWR.beamline.online_processing is not None
                ):
                    self.online_processing_task = gevent.spawn(
                        HWR.beamline.online_processing.run_processing, dc
                    )

                empty_cpos = queue_model_objects.CentredPosition()
                if cpos != empty_cpos:
                    HWR.beamline.sample_view.select_shape_with_cpos(cpos)
                else:
                    pos_dict = HWR.beamline.diffractometer.get_positions()
                    cpos = queue_model_objects.CentredPosition(pos_dict)
                    snapshot = HWR.beamline.sample_view.get_snapshot()
                    acq_1.acquisition_parameters.centred_position = cpos
                    acq_1.acquisition_parameters.centred_position.snapshot_image = (
                        snapshot
                    )

                # inc_used_for_collection does nothing
                HWR.beamline.sample_view.inc_used_for_collection(cpos)
                param_list = queue_model_objects.to_collect_dict(
                    dc,
                    sample,
                    cpos if cpos != empty_cpos else None,
                )

                # TODO this is wrong. Rename to something like collect.start_procedure
                self.collect_task = HWR.beamline.collect.collect(
                    COLLECTION_ORIGIN_STR.MXCUBE, param_list
                )
                self.collect_task.get()

                if "collection_id" in param_list[0]:
                    dc.id = param_list[0]["collection_id"]

                dc.acquisitions[0].path_template.xds_dir = param_list[0]["xds_dir"]

            except gevent.GreenletExit:
                # log.warning("Collection stopped by user.")
                list_item.setText(1, "Stopped")
                raise QueueAbortedException("queue stopped by user", self)
            except Exception as ex:
                raise QueueExecutionException(str(ex), self)
        else:
            log.error(
                "Could not call the data collection routine,"
                + " check the beamline configuration"
            )
            list_item.setText(1, "Failed")
            msg = (
                "Could not call the data collection"
                + " routine, check the beamline configuration"
            )
            raise QueueExecutionException(msg, self)

    def collect_started(self, owner, num_oscillations):
        logging.getLogger("user_level_log").info("Collection started")
        self.get_view().setText(1, "Collecting...")

    def collect_number_of_frames(self, number_of_images=0, exposure_time=0):
        pass

    def image_taken(self, image_number):
        if image_number > 0:
            num_images = (
                self.get_data_model().acquisitions[0].acquisition_parameters.num_images
            )
            num_images += (
                self.get_data_model().acquisitions[0].acquisition_parameters.first_image
                - 1
            )
            self.get_view().setText(1, str(image_number) + "/" + str(num_images))

    def preparing_collect(self, number_images=0, exposure_time=0):
        self.get_view().setText(1, "Preparing to collecting")

    def collect_failed(self, owner, state, message, *args):
        # this is to work around the remote access problem
        dispatcher.send("collect_finished")
        self.get_view().setText(1, "Failed")
        self.status = QUEUE_ENTRY_STATUS.FAILED
        logging.getLogger("queue_exec").error(message.replace("\n", ";  "))
        # raise QueueExecutionException(message.replace("\n", " "), self)

    def collect_osc_started(
        self, owner, blsampleid, barcode, location, collect_dict, osc_id
    ):
        self.get_view().setText(1, "Preparing")

    def collect_finished(self, owner, state, message, *args):
        # this is to work around the remote access problem
        dispatcher.send("collect_finished")
        self.get_view().setText(1, "Collection done")
        logging.getLogger("user_level_log").info("Collection finished")

        if self.online_processing_task is not None:
            self.get_view().setText(1, "Processing...")
            logging.getLogger("user_level_log").warning("Processing: Please wait...")
            HWR.beamline.online_processing.done_event.wait(timeout=120)
            HWR.beamline.online_processing.done_event.clear()

    def stop(self):
        BaseQueueEntry.stop(self)
        HWR.beamline.collect.stop_collect()
        if self.online_processing_task is not None:
            HWR.beamline.online_processing.stop_processing()
            logging.getLogger("user_level_log").error("Processing: Stopped")
        if self.centring_task is not None:
            self.centring_task.kill(block=False)

        self.get_view().setText(1, "Stopped")
        logging.getLogger("queue_exec").info("Calling stop on: " + str(self))
        logging.getLogger("user_level_log").info("Collection stopped")
        # this is to work around the remote access problem
        dispatcher.send("collect_finished")
        raise QueueAbortedException("Queue stopped", self)

    def online_processing_finished(self):
        dispatcher.send("collect_finished")
        self.online_processing_task = None
        # self.get_view().setText(1, "Done")
        logging.getLogger("user_level_log").info("Processing: Done")

    def online_processing_failed(self):
        self.online_processing_task = None
        self.get_view().setText(1, "Processing failed")
        logging.getLogger("user_level_log").error("Processing: Failed")

    def get_type_str(self):
        data_model = self.get_data_model()
        if data_model.is_helical():
            return "Helical"
        elif data_model.is_mesh():
            return "Mesh"
        else:
            return "OSC"

    def add_processing_msg(self, time, method, status, msg):
        data_model = self.get_data_model()
        data_model.add_processing_msg(time, method, status, msg)
        self.get_view().update_tool_tip()



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging

import gevent
import gevent.event

try:
    import Queue as queue
except ImportError:
    import queue

import numpy

from mxcubecore import Poller
from mxcubecore.CommandContainer import (
    ChannelObject,
    CommandObject,
    ConnectionError,
)
from mxcubecore.dispatcher import saferef

gevent_version = list(map(int, gevent.__version__.split(".")))

log = logging.getLogger("HWR")

try:
    import tango
    from tango.gevent import DeviceProxy
except ImportError:
    logging.getLogger("HWR").warning("Tango support is not available.")


__copyright__ = """ Copyright © 2010 - 2020 by MXCuBE Collaboration """
__license__ = "LGPLv3+"

log = logging.getLogger("HWR")


class TangoCommand(CommandObject):
    def __init__(self, name, command, tangoname=None, username=None, **kwargs):
        CommandObject.__init__(self, name, username, **kwargs)

        self.command = command
        self.device_name = tangoname
        self.device = None

    def init_device(self):
        try:
            self.device = DeviceProxy(self.device_name)
        except tango.DevFailed:
            logging.getLogger("HWR").exception(
                "Failed to set-up tango command %s on %s device",
                self.name(),
                self.device_name,
            )
            self.device = None
        else:
            try:
                self.device.ping()
            except tango.ConnectionFailed:
                self.device = None
                raise ConnectionError

    def __call__(self, *args, **kwargs):
        self.emit("commandBeginWaitReply", (str(self.name()),))
        if self.device is None:
            # TODO: emit commandFailed
            # beware of infinite recursion with Sample Changer
            # (because of procedure exception cleanup...)
            self.init_device()

        try:
            tango_cmd_object = getattr(self.device, self.command)
            ret = tango_cmd_object(
                *args
            )  # eval('self.device.%s(*%s)' % (self.command, args))
        except tango.DevFailed:
            logging.getLogger("HWR").exception(
                "Failed to run tango command %s on %s device",
                self.name(),
                self.device_name,
            )
        except Exception:
            logging.getLogger("HWR").exception(
                "%s: an error occured when calling Tango command %s",
                str(self.name()),
                self.command,
            )
        else:
            self.emit("commandReplyArrived", (ret, str(self.name())))
            return ret
        self.emit("commandFailed", (-1, self.name()))

    def abort(self):
        pass

    def set_device_timeout(self, timeout):
        if self.device is None:
            self.init_device()
        self.device.set_timeout_millis(timeout)

    def is_connected(self):
        return self.device is not None


def process_tango_events():
    while not TangoChannel._tangoEventsQueue.empty():
        try:
            ev = TangoChannel._tangoEventsQueue.get_nowait()
        except queue.Empty:
            break
        else:
            try:
                receiverCbRef = TangoChannel._eventReceivers[id(ev)]
                receiverCb = receiverCbRef()
                if receiverCb is not None:
                    try:
                        gevent.spawn(receiverCb, ev.event.attr_value.value)
                    except AttributeError:
                        logging.getLogger("HWR").exception("")
            except KeyError:
                logging.getLogger("HWR").exception("")


class E:
    def __init__(self, event):
        self.event = event


def _device_has_attribute(device: DeviceProxy, attribute_name: str) -> bool:
    """Check if a tango device has an attribute."""

    try:
        device.attribute_query(attribute_name)
    except tango.DevFailed as ex:
        if ex.args[0].reason == "API_AttrNotFound":
            # query failed with 'attribute not found' error
            return False

        # unexpected exception, re-raise
        raise

    # attribute query was successful, thus we know the attribute exits
    return True


class TangoChannel(ChannelObject):
    _tangoEventsQueue = queue.Queue()
    _eventReceivers = {}

    _tangoEventsProcessingTimer = gevent.get_hub().loop.async_()

    # start Tango events processing timer
    _tangoEventsProcessingTimer.start(process_tango_events)

    def __init__(
        self,
        name,
        attribute_name,
        tangoname=None,
        username=None,
        polling=None,
        timeout=10000,
        **kwargs,
    ):
        ChannelObject.__init__(self, name, username, **kwargs)

        self.attribute_name = attribute_name
        self.device_name = tangoname
        self.device = None
        self.value = Poller.NotInitializedValue
        self.poller = None
        self.polling = polling
        self.polling_timer = None
        self.polling_events = False
        self.timeout = int(timeout)
        self.read_as_str = kwargs.get("read_as_str", False)
        self._device_initialized = gevent.event.Event()
        self.init_device()
        self.continue_init(None)

    def stop_polling(self):
        """Stop polling the underlying Tango attribute.

        If this channel is currently polling its tango attribute, via
        'attribute read' calls, stop polling.

        If no polling is active, this method does nothing.
        """
        if self.poller is not None:
            self.poller.stop()
            self.poller = None

    def init_poll_failed(self, e, poller_id):
        self._device_initialized.clear()
        logging.warning(
            "%s/%s (%s): could not complete init. "
            "(hint: device server is not running, or has to be restarted)",
            self.device_name,
            self.attribute_name,
            self.name(),
        )
        self.init_poller = self.init_poller.restart(3000)

    def continue_init(self, _):
        if isinstance(self.polling, int):
            self.raw_device = DeviceProxy(self.device_name)

            self.poller = Poller.poll(
                self.poll,
                polling_period=self.polling,
                value_changed_callback=self.update,
                error_callback=self.poll_failed,
            )
        else:
            if self.polling == "events":
                # try to register event
                try:
                    self.polling_events = True
                    self.device.subscribe_event(
                        self.attribute_name,
                        tango.EventType.CHANGE_EVENT,
                        self,
                        [],
                        True,
                    )
                except Exception:
                    logging.getLogger("HWR").exception("could not subscribe event")
        self._device_initialized.set()

    def init_device(self):
        try:
            self.device = DeviceProxy(self.device_name)
        except tango.DevFailed:
            self.imported = False
            logging.getLogger("HWR").exception(
                "Failed to set-up tango channel %s on %s device",
                self.name(),
                self.device_name,
            )
        else:
            self.imported = True
            try:
                self.device.ping()
            except tango.ConnectionFailed:
                self.device = None
                raise ConnectionError
            else:
                self.device.set_timeout_millis(self.timeout)

                # check that the attribute exists (to avoid Abort in PyTango grrr)
                if not _device_has_attribute(self.device, self.attribute_name):
                    logging.getLogger("HWR").error(
                        "no attribute %s in Tango device %s",
                        self.attribute_name,
                        self.device_name,
                    )
                    self.device = None

    def push_event(self, event):
        if (
            event.attr_value is None
            or event.err
            or event.attr_value.quality != tango.AttrQuality.ATTR_VALID
        ):
            return

        ev = E(event)
        TangoChannel._eventReceivers[id(ev)] = saferef.safe_ref(self.update)
        TangoChannel._tangoEventsQueue.put(ev)
        TangoChannel._tangoEventsProcessingTimer.send()

    def poll(self):
        if self.read_as_str:
            value = self.raw_device.read_attribute(
                self.attribute_name, tango.DeviceAttribute.ExtractAs.String
            ).value
        else:
            value = self.raw_device.read_attribute(self.attribute_name).value

        return value

    def poll_failed(self, e, poller_id):
        self.emit("update", None)
        poller = Poller.get_poller(poller_id)
        if poller is not None:
            poller.restart(1000)

    def get_info(self):
        self._device_initialized.wait(timeout=3)
        return self.device.get_attribute_config(self.attribute_name)

    def update(self, value=Poller.NotInitializedValue):
        # start with checking if we have a numpy array, as comparing
        # numpy.ndarray to Poller.NotInitializedValue raises a ValueError exception
        if isinstance(value, numpy.ndarray):
            value = value.tolist()
        elif value == Poller.NotInitializedValue:
            value = self.get_value()
        elif isinstance(value, tuple):
            value = list(value)

        self.value = value
        self.emit("update", value)

    def get_value(self):
        if self.read_as_str:
            value = self.device.read_attribute(
                self.attribute_name, tango.DeviceAttribute.ExtractAs.String
            ).value
        else:
            value = self.device.read_attribute(self.attribute_name).value

        if isinstance(value, numpy.ndarray):
            if not numpy.array_equal(value, self.value):
                self.update(value)
        elif value != self.value:
            self.update(value)

        return value

    def set_value(self, new_value):
        self.device.write_attribute(self.attribute_name, new_value)

    def is_connected(self):
        return self.device is not None



"""
Combine setting the back light IN with moving the beamstop out.
This is used to set the back light in/out faster, than using
the phase CENTRING - the beamstop motor moves only if needed.
Example xml file:
<object class="MicrodiffLightBeamstop">
  <username>Back Light</username>
  <exporter_address>wid30bmd2s:9001</exporter_address>
  <actuator_name>BackLightIsOn</actuator_name>
  <values>{"IN": True, "OUT": False}</values>
  <use_hwstate>True</use_hwstate>
  <object role="beamstop" href="/udiff_beamstop"/>
</object>
"""

from mxcubecore.HardwareObjects.ExporterNState import ExporterNState

__copyright__ = """ Copyright © 2010-2020 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class MicrodiffLightBeamstop(ExporterNState):
    """Control backlight, move the beamstop to safety position"""

    def __init__(self, name):
        super().__init__(name)
        self._beamstop_obj = None
        self._saved_beamstop_value = None

    def init(self):
        """Initialize the light and the beamstop object"""
        super().init()

        # get the beamstop object
        self._beamstop_obj = self.get_object_by_role("beamstop")

    def _set_value(self, value):
        """Set light to value. Move the beamstop, if needed.
        Args:
            value (enum): Value to be set.
        """
        if value == self.VALUES.IN:
            # move the beamstop out before setting the back light in
            if self._beamstop_obj:
                self.handle_beamstop(value)
            super()._set_value(value)
        else:
            super()._set_value(value)
            # move the beamstop if needed after getting the light out
            if self._beamstop_obj:
                self.handle_beamstop(value)

    def handle_beamstop(self, value):
        """Move the beamstop as function of the value of the back light.
        Args:
            (enum): light value.
        """
        if value == self.VALUES.IN:
            self._saved_beamstop_value = self._beamstop_obj.get_value()
            # move the beamstop out to avoid collision with the back light
            if self._saved_beamstop_value == self._beamstop_obj.VALUES.IN:
                self._beamstop_obj.set_value(self._beamstop_obj.VALUES.OUT, timeout=60)
        else:
            if self._saved_beamstop_value:
                self._beamstop_obj.set_value(self._saved_beamstop_value, timeout=60)
                self._saved_beamstop_value = None



from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject


class PX1Configuration(HardwareObject):
    def init(self):
        self.use_edna_value = self.get_property("use_edna")
        self.pin_length = self.get_property("pin_length")

        self.centring_points = self.get_property("centring_points")
        self.centring_phi_incr = self.get_property("centring_phi_increment")
        self.centring_sample_type = self.get_property("centring_sample_type")

        print("LocalConfiguration has value sample_type=%s" % self.centring_sample_type)

    def getUseEDNA(self):
        return self.use_edna_value

    def setUseEDNA(self, value):
        if value is True or value == "True":
            self.use_edna_value = True
            self.setProperty("use_edna", True)
        else:
            self.use_edna_value = False
        self.setProperty("use_edna", self.use_edna_value)

    def getPinLength(self):
        return self.pin_length

    def setPinLength(self, value):
        self.pin_length = value
        self.setProperty("pin_length", value)

    def getCentringPoints(self):
        return int(self.centring_points)

    def setCentringPoints(self, value):
        self.centring_points = int(value)
        self.setProperty("centring_points", value)

    def getCentringPhiIncrement(self):
        return float(self.centring_phi_incr)

    def setCentringPhiIncrement(self, value):
        self.centring_phi_incr = float(value)
        self.setProperty("centring_phi_increment", value)

    def getCentringSampleType(self):
        return self.centring_sample_type

    def setCentringSampleType(self, value):
        self.centring_sample_type = value
        self.setProperty("centring_sample_type", value)

    def save(self):
        self.commit_changes()


if __name__ == "__main__":
    hwr = HWR.get_hardware_repository()
    hwr.connect()

    env = hwr.get_hardware_object("/px1configuration")

    print("PX1 Configuration ")
    use_edna = env.getUseEDNA()
    print("    use_edna %s / (type: %s)" % (use_edna, type(use_edna)))
    print("    pin_length", env.getPinLength())
    print("    centring")
    print("       nb points", env.getCentringPoints())
    print("       phi incr", env.getCentringPhiIncrement())
    print("       sample type", env.getCentringSampleType())
    env.setUseEDNA("False")
    env.setPinLength("10")
    print("    use_edna %s " % env.getUseEDNA())
    print("    pin_length", env.getPinLength())
    # env.save()
